/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.document.dtd;

import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;

import org.w3c.dom.Element;

import org.xmloperator.framework.document.dtd.SpecificationNode;
import org.xmloperator.framework.document.schema.ElementContent;
import org.xmloperator.framework.document.schema.
  NotConformantToContentModelException;
import org.xmloperator.framework.document.schema.UnsatisfiedPatternException;
import org.xmloperator.framework.document.schema.
    InvalidElementSpecificationException;
import org.xmloperator.document.dtd.SpecNodeUtils;

public final class ElementContentImpl implements ElementContent {

  public static final boolean isContentValid(
    final SpecificationNode rootSpecNode, final String[] contentElementNames) {
    try {
      new ElementContentImpl(rootSpecNode, contentElementNames);
      return true;
    }
    catch (NotConformantToContentModelException exception) {
      return false;
    }
  }

  private SpecificationNode rootSpecNode;
  private int deepth;
  private String[] contentElementNames;
  private int[][] elementNamePaths;

  public ElementContentImpl(final SpecificationNode rootSpecNode,
    final String[] contentElementNames)
    throws NotConformantToContentModelException {
    if (rootSpecNode == null || contentElementNames == null)
      throw new IllegalArgumentException();
    for (int k = 0; k < contentElementNames.length; k++)
      if (contentElementNames[k] == null)
        throw new IllegalArgumentException();
    this.rootSpecNode = rootSpecNode;
    this.deepth = SpecNodeUtils.computeDeepth(rootSpecNode);
    this.contentElementNames = contentElementNames;
    int elementNameCount = contentElementNames.length;
    final ElementNamePathComputer elementNamePathComputer =
      new ElementNamePathComputer();
    this.elementNamePaths = new int[elementNameCount][deepth + 1];
    for (int k = 0; k < elementNameCount; k++)
      try {
        int[] elementNamePath = elementNamePaths[k];
        Arrays.fill(elementNamePath, -1);
        elementNamePathComputer.setElementSpec(contentElementNames[k],
            elementNamePath, k != 0 ? elementNamePaths[k - 1] : null);
        elementNamePathComputer.computeElementNamePath(rootSpecNode);
      }
      catch (NotConformantToContentModelException exception) {
        exception.contentElementSpecIndex = k;
        throw exception;
      }
    try { // is there an unsatisfied pattern ?
      int[] elementNamePath = new int[deepth + 1];
      Arrays.fill(elementNamePath, -1);
      elementNamePathComputer.setElementSpec("", elementNamePath,
        elementNameCount != 0 ? elementNamePaths[elementNameCount - 1] : null);
      elementNamePathComputer.computeElementNamePath(rootSpecNode);
    }
    catch (InvalidElementSpecificationException exception) {
      // Expected exception. "" is not a valid name. No unsatisfied pattern.
    }
  }

  private static class ElementNamePathComputer {
    private static final InvalidElementSpecificationException
      INVALID_ELEMENT_SPECIFICATION_EXCEPTION =
      new InvalidElementSpecificationException();

    private String elementName;
    private int[] elementNamePath;
    private int[] previousElementNamePath; // may be null
    private final List unsatisfiedPatternList = new ArrayList();

    public ElementNamePathComputer() {
    }

    public void setElementSpec(String elementName,
      int[] elementNamePath, int[] previousElementNamePath) {
      this.elementName = elementName;
      this.elementNamePath = elementNamePath;
      this.previousElementNamePath = previousElementNamePath;
    }

    private UnsatisfiedPatternException createUnsatisfiedPatternException() {
      String[] unsatisfiedPatterns = new String[unsatisfiedPatternList.size()];
      unsatisfiedPatternList.toArray(unsatisfiedPatterns);
      return new UnsatisfiedPatternException(unsatisfiedPatterns);
    }

    // Compute elementNamePath from rootSpectNode for succeeding
    //   to previousElementNamePath.
    public void computeElementNamePath(final SpecificationNode rootSpecNode)
      throws NotConformantToContentModelException {
      unsatisfiedPatternList.clear();
      try {
        if (previousElementNamePath != null) {
          computeElementNamePathSucceeding(rootSpecNode, 0, false);
          throw INVALID_ELEMENT_SPECIFICATION_EXCEPTION;
        }
        else {
          if (computeElementNamePath(rootSpecNode, 0))
            throw createUnsatisfiedPatternException();
          else
            throw INVALID_ELEMENT_SPECIFICATION_EXCEPTION;
        }
      }
      catch (ElementNameFoundException exception) {
        // Expected exception.
      }
    }

    // Compute the element name path that is the first
    //   to succeed to the previous one and to map to the element name.
    // Throws an ElementNameFoundException if succeeds
    //   or raises an UnsatisfiedPatternException
    //   if an element name is unsatisfied.
    private void computeElementNamePathSucceeding(
      final SpecificationNode specNode, final int startingIndex,
      boolean canRepeatPreviousPath)
      throws ElementNameFoundException, UnsatisfiedPatternException {
      if (!canRepeatPreviousPath)
        canRepeatPreviousPath =
          SpecNodeUtils.isOccurrenceRepeatable(specNode.getOccurrence());
      int childCount = specNode.getChildCount();
      if (childCount == 0) {
        if (!canRepeatPreviousPath
          || !elementName.equals(specNode.getElementName()))
          return;
        elementNamePath[startingIndex] = -1;
        throw new ElementNameFoundException();
      }
      int previousPathStep = previousElementNamePath[startingIndex];
      elementNamePath[startingIndex] = previousPathStep;
      boolean isSequence = specNode.isSequence();
      computeElementNamePathSucceeding(
        specNode.getChild(previousPathStep), startingIndex + 1,
        canRepeatPreviousPath && !isSequence);
      if (isSequence)
        for (int k = previousPathStep + 1; k < childCount; k++) {
          elementNamePath[startingIndex] = k;
          SpecificationNode childSpecNode = specNode.getChild(k);
          if (computeElementNamePath(childSpecNode, startingIndex + 1))
            throw createUnsatisfiedPatternException();
        }
      if (canRepeatPreviousPath)
        computeElementNamePath(specNode, startingIndex);
    }

    // Compute the element name path that is the first
    //   to map to the element name.
    // Throws an ElementNameFoundException if succeeds
    //   elsewhere return true if an element name is unsatisfied
    private boolean computeElementNamePath(final SpecificationNode specNode,
      final int startingIndex) throws ElementNameFoundException {
      boolean isOccurrenceRequired =
        SpecNodeUtils.isOccurrenceRequired(specNode.getOccurrence());
      int childCount = specNode.getChildCount();
      if (childCount == 0) {
        String leafName = specNode.getElementName();
        if (leafName == null)
          return false;
        if (!elementName.equals(leafName)) {
          if (isOccurrenceRequired)
            unsatisfiedPatternList.add(leafName);
          return isOccurrenceRequired;
        }
        elementNamePath[startingIndex] = -1;
        throw new ElementNameFoundException();
      }
      if (specNode.isSequence()) {
        for (int k = 0; k < childCount; k++) {
          elementNamePath[startingIndex] = k;
          if (computeElementNamePath(specNode.getChild(k), startingIndex + 1))
            return isOccurrenceRequired;
        }
        return false;
      }
      else {
        int unsatisfiedPatternCount = unsatisfiedPatternList.size();
        boolean isPatternUnsatisfied = isOccurrenceRequired;
        for (int k = 0; k < childCount; k++) {
          elementNamePath[startingIndex] = k;
          isPatternUnsatisfied &=
            computeElementNamePath(specNode.getChild(k), startingIndex + 1);
        }
        if (!isPatternUnsatisfied)
          truncateList(unsatisfiedPatternList, unsatisfiedPatternCount);
        return isPatternUnsatisfied;
      }
    }

    private static final void truncateList(List list, int count) {
      while (true) {
        int size = list.size();
        if (size <= count)
          return;
        list.remove(size - 1);
      }
    }

    private static class ElementNameFoundException extends Exception {

      public ElementNameFoundException() {
      }
    }
  }

  ////////////////////////////////
  // ElementContent implementation
  ////////////////////////////////

  public int getElementSpecCount() {
    return contentElementNames.length;
  }

  // Return an element name string.
  public Object getElementSpec(final int contentElementSpecIndex) {
    return contentElementNames[contentElementSpecIndex];
  }

  // Return an element name string.
  public Object getElementDefinition(final int contentElementSpecIndex) {
    return contentElementNames[contentElementSpecIndex];
  }

  public boolean isElementDefinitionRepeatable(
      final int contentElementSpecIndex) {
    return SpecNodeUtils.isElementNameRepeatable(rootSpecNode,
      elementNamePaths[contentElementSpecIndex]);
  }

  public void addElementDefinitionsForReplace(final Set contentElementSpecSet,
    final int contentElementSpecIndex) {
    SpecNodeUtils.addElementNamesForReplace(contentElementSpecSet,
        rootSpecNode, elementNamePaths[contentElementSpecIndex]);
  }

  public Set elementDefinitionsInsertableBefore(
      final int contentElementSpecIndex) {
    Set elementNameSet = new HashSet();
    boolean isElementBefore = contentElementSpecIndex > 0;
    boolean isElementAfter =
        contentElementSpecIndex < contentElementNames.length;
    int[] elementNamePath1 =
      isElementBefore ? elementNamePaths[contentElementSpecIndex - 1] : null;
    int[] elementNamePath2 =
      isElementAfter ? elementNamePaths[contentElementSpecIndex] : null;
    if (isElementBefore && SpecNodeUtils.isElementNameRepeatable(rootSpecNode,
      elementNamePath1)) {
      elementNameSet.add(contentElementNames[contentElementSpecIndex - 1]);
      SpecNodeUtils.addElementNamesForReplaceRepeatedUse(
          elementNameSet, rootSpecNode, elementNamePath1);
    }
    if (isElementAfter && SpecNodeUtils.isElementNameRepeatable(rootSpecNode,
      elementNamePath2)) {
      elementNameSet.add(contentElementNames[contentElementSpecIndex]);
      SpecNodeUtils.addElementNamesForReplaceRepeatedUse(
          elementNameSet, rootSpecNode, elementNamePath2);
    }
    SpecNodeUtils.addElementNamesInsertableBetween(elementNameSet,
      rootSpecNode, elementNamePath1, elementNamePath2);
    return elementNameSet;
  }

  public String elementFromContentModelToString(
      final int contentElementSpecIndex, final boolean isSummarized,
      final Element contextElement) {
    return SpecNodeUtils.specNodeToString(rootSpecNode,
      elementNamePaths[contentElementSpecIndex], isSummarized);
  }
}
