/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.document.dtd;

import java.util.Set;

import org.xmloperator.framework.document.dtd.SpecificationNode;

public abstract class SpecNodeUtils {

  ///////////////////////////////////////////////////////////////////
  // Methods for geting general information about a SpecificationNode
  ///////////////////////////////////////////////////////////////////

  // return the specification node string as it can appear in the DTD
  public static final String specNodeToString(final SpecificationNode specNode)
    {
    final StringBuffer s = new StringBuffer();
    appendSpecNodeToString(s, true, specNode);
    return s.toString();
  }

  private static final void appendSpecNodeToString(final StringBuffer s,
    final boolean isRoot, final SpecificationNode specNode) {
    if (isRoot)
      s.append('(');
    final int childCount = specNode.getChildCount();
    if (childCount == 0) {
      String elementName = specNode.getElementName();
      s.append(elementName != null ? elementName : "#PCDATA");
    }
    else {
      if (!isRoot)
        s.append('(');
      final char separator = specNode.isSequence() ? ',' : '|';
      for (int k = 0; k < childCount; k++) {
        if (k > 0)
          s.append(separator);
        appendSpecNodeToString(s, false, specNode.getChild(k));
      }
      if (!isRoot)
        s.append(')');
    }
    if (isRoot)
      s.append(')');
    char occurrence = specNode.getOccurrence();
    if (occurrence != '.')
      s.append(occurrence);
  }

  // return the maximun length of element name path in a SpecificationNode
  public static final int computeDeepth(final SpecificationNode specNode) {
    int maxDeepth = 0;
    for (int k = 0; k < specNode.getChildCount(); k++) {
      int deepth = 1 + computeDeepth(specNode.getChild(k));
      if (deepth > maxDeepth)
        maxDeepth = deepth;
    }
    return maxDeepth;
  }

  ////////////////////////////
  // Methods about occurrences
  ////////////////////////////

  public static final boolean isOccurrenceRequired(final char occurrence) {
    return occurrence != '?' && occurrence != '*';
  }

  public static final boolean isOccurrenceRepeatable(final char occurrence) {
    return occurrence == '+' || occurrence == '*';
  }

  //////////////////////////////////////////////////////////////
  // Methods about having required or not required element names
  //////////////////////////////////////////////////////////////

  // Return true if a SpecificationNode has at least an element name
  public static final boolean hasElementNames(final SpecificationNode specNode)
    {
    return specNode.getChildCount() != 0 || specNode.getElementName() != null;
  }

  // Returns true if one element name at least is required
  public static final boolean hasRequiredElementNames(
    final SpecificationNode specNode) {
    if (!isOccurrenceRequired(specNode.getOccurrence()))
      return false;
    final int childCount = specNode.getChildCount();
    if (childCount == 0)
      return specNode.getElementName() != null;
    if (specNode.isSequence()) {
      for (int k = 0; k < childCount; k++)
        if (hasRequiredElementNames(specNode.getChild(k)))
          return true;
      return false;
    }
    else {
      for (int k = 0; k < childCount; k++)
        if (!hasRequiredElementNames(specNode.getChild(k)))
          return false;
      return true;
    }
  }

  /////////////////////////////////////////////////////////////////////
  // Methods about geting the required or not required element name set
  //   of a SpecificationNode
  /////////////////////////////////////////////////////////////////////

  // Adds all element names of a SpecificationNode to an element name set.
  // Returns true if one element name at least is required.
  public static final boolean addElementNames(final Set elementNameSet,
    final SpecificationNode specNode) {
    final boolean isOccurrenceRequired =
      isOccurrenceRequired(specNode.getOccurrence());
    final int childCount = specNode.getChildCount();
    if (childCount == 0) {
      final String elementName = specNode.getElementName();
      if (elementName == null)
        return false;
      elementNameSet.add(elementName);
      return isOccurrenceRequired;
    }
    boolean hasRequiredElementNames;
    if (specNode.isSequence()) {
      hasRequiredElementNames = false;
      for (int k = 0; k < childCount; k++)
        hasRequiredElementNames |=
          addElementNames(elementNameSet, specNode.getChild(k));
    }
    else {
      hasRequiredElementNames = true;
      for (int k = 0; k < childCount; k++)
        hasRequiredElementNames &=
          addElementNames(elementNameSet, specNode.getChild(k));
    }
    return hasRequiredElementNames & isOccurrenceRequired;
  }

  // Adds all first element names of sequences in a SpecificationNode
  //   to an element name set.
  // Returns true if one element name at least is required
  public static final boolean addFirstElementNames(final Set elementNameSet,
    final SpecificationNode specNode) {
    final boolean isOccurrenceRequired =
      isOccurrenceRequired(specNode.getOccurrence());
    final int childCount = specNode.getChildCount();
    if (childCount == 0) {
      String elementName = specNode.getElementName();
      if (elementName == null)
        return false;
      elementNameSet.add(elementName);
      return isOccurrenceRequired;
    }
    if (specNode.isSequence()) {
      for (int k = 0; k < childCount; k++)
        if (addFirstElementNames(elementNameSet, specNode.getChild(k)))
          return isOccurrenceRequired;
      return false;
    }
    else {
      boolean hasRequiredElementNames = true;
      for (int k = 0; k < childCount; k++)
        hasRequiredElementNames &=
          addFirstElementNames(elementNameSet, specNode.getChild(k));
      return hasRequiredElementNames & isOccurrenceRequired;
    }
  }

  // Adds all required element names of a SpecificationNode
  //   to an element name set.
  // Returns true if one element name at least is required
  public static final boolean addRequiredElementNames(final Set elementNameSet,
    SpecificationNode specNode) {
    if (!isOccurrenceRequired(specNode.getOccurrence()))
      return false;
    final int childCount = specNode.getChildCount();
    if (childCount == 0) {
      final String elementName = specNode.getElementName();
      if (elementName == null)
        return false;
      elementNameSet.add(elementName);
      return true;
    }
    if (specNode.isSequence()) {
      boolean hasRequiredElementNames = false;
      for (int k = 0; k < childCount; k++)
        hasRequiredElementNames |=
          addRequiredElementNames(elementNameSet, specNode.getChild(k));
      return hasRequiredElementNames;
    }
    else {
      for (int k = 0; k < childCount; k++)
        if (!hasRequiredElementNames(specNode.getChild(k)))
          return false;
      for (int k = 0; k < childCount; k++)
        addRequiredElementNames(elementNameSet, specNode.getChild(k));
      return true;
    }
  }

  // Adds all first required element names of sequences in a SpecificationNode
  //   to an element name set.
  // Returns true if one element name at least is required
  public static final boolean addFirstRequiredElementNames(
    final Set elementNameSet, SpecificationNode specNode) {
    if (!isOccurrenceRequired(specNode.getOccurrence()))
      return false;
    final int childCount = specNode.getChildCount();
    if (childCount == 0) {
      final String elementName = specNode.getElementName();
      if (elementName == null)
        return false;
      elementNameSet.add(elementName);
      return true;
    }
    if (specNode.isSequence()) {
      for (int k = 0; k < childCount; k++)
        if (addFirstRequiredElementNames(elementNameSet, specNode.getChild(k)))
          return true;
      return false;
    }
    else {
      for (int k = 0; k < childCount; k++)
        if (!hasRequiredElementNames(specNode.getChild(k)))
          return false;
      for (int k = 0; k < childCount; k++)
        addFirstRequiredElementNames(elementNameSet, specNode.getChild(k));
      return true;
    }
  }

  ///////////////////////////////////
  // Methods about element name paths
  ///////////////////////////////////
  // An element name path is a sequence of index terminated by -1.
  ////////////////////////////////////////////////////////////////

  // return the specification node string focused on a node.
  public static final String specNodeToString(final SpecificationNode specNode,
    int[] elementNamePath, boolean isSummarized) {
    final StringBuffer s = new StringBuffer();
    appendSpecNodeToString(s, true, specNode, elementNamePath, 0, isSummarized);
    return s.toString();
  }

  private static final void appendSpecNodeToString(final StringBuffer s,
    final boolean isRoot, final SpecificationNode specNode,
    final int[] elementNamePath, final int startIndex,
    final boolean isSummarized) {
    if (isRoot)
      s.append('(');
    final int childCount = specNode.getChildCount();
    if (childCount == 0) {
      final String space = elementNamePath != null ? "    " : "";
      s.append(space);
      String elementName = specNode.getElementName();
      s.append(elementName != null ? elementName : "#PCDATA");
      s.append(space);
    }
    else {
      if (!isRoot)
        s.append('(');
      final int elementNamePathStep =
        elementNamePath != null ? elementNamePath[startIndex] : -1;
      final String separator = specNode.isSequence() ? " , " : " | ";
      if (isSummarized && childCount > 2) {
        appendSpecNodeToString(s, false, specNode.getChild(0),
          elementNamePathStep == 0 ? elementNamePath : null, startIndex + 1,
          isSummarized);
        s.append(separator);
        appendSpecNodeToString(s, false, specNode.getChild(1),
          elementNamePathStep == 1 ? elementNamePath : null, startIndex + 1,
          isSummarized);
        final String summarizedGroup = " ... ";
        if (elementNamePathStep != 2) {
          s.append(separator);
          s.append(summarizedGroup);
        }
        if (elementNamePathStep >= 2) {
          s.append(separator);
          appendSpecNodeToString(s, false,
            specNode.getChild(elementNamePathStep), elementNamePath,
            startIndex + 1, isSummarized);
          if (elementNamePathStep < childCount - 1) {
            s.append(separator);
            s.append(summarizedGroup);
          }
        }
      }
      else
        for (int k = 0; k < childCount; k++) {
          if (k > 0)
            s.append(separator);
          appendSpecNodeToString(s, false, specNode.getChild(k),
            elementNamePathStep == k ? elementNamePath : null, startIndex + 1,
            isSummarized);
        }
      if (!isRoot)
        s.append(')');
    }
    if (isRoot)
      s.append(')');
    final char occurrence = specNode.getOccurrence();
    if (occurrence != '.')
      s.append(occurrence);
  }

  // Return the SpecificationNode at the end of an element name path
  public static final SpecificationNode getSpecificationNodeFromPath(
    final SpecificationNode rootSpecNode, final int[] elementNamePath) {
    SpecificationNode specNode = rootSpecNode;
    int pathIndex = 0;
    while (true) {
      int childIndex = elementNamePath[pathIndex];
      if (childIndex < 0)
        return specNode;
      if (specNode.getChildCount() == 0)
        throw new IllegalArgumentException();
      specNode = specNode.getChild(childIndex);
      pathIndex++;
    }
  }

  // Return the element name (or null) at the end of an element name path
  public static final String getElementNameFromPath(
    final SpecificationNode rootSpecNode, final int[] elementNamePath) {
    SpecificationNode specNode = rootSpecNode;
    int pathIndex = 0;
    while (true) {
      int childIndex = elementNamePath[pathIndex];
      boolean isElementNamePathEnd = childIndex < 0;
      if ((specNode.getChildCount() == 0) != isElementNamePathEnd)
        // The element name path end must map to a leaf spec node
        throw new IllegalArgumentException();
      if (isElementNamePathEnd)
        return specNode.getElementName();
      specNode = specNode.getChild(childIndex);
      pathIndex++;
    }
  }

  // Return true if the element name referenced by an element name path
  //   may be repeated.
  public static final boolean isElementNameRepeatable(
    final SpecificationNode rootSpecNode, final int[] elementNamePath) {
    SpecificationNode specNode = rootSpecNode;
    boolean isRepeatable = false;
    int pathIndex = 0;
    while (true) {
      if (!isRepeatable)
        isRepeatable = isOccurrenceRepeatable(specNode.getOccurrence());
      int childIndex = elementNamePath[pathIndex];
      boolean isElementNamePathEnd = childIndex < 0;
      if ((specNode.getChildCount() == 0) != isElementNamePathEnd)
        // The element name path end must map to a leaf spec node
        throw new IllegalArgumentException();
      if (isElementNamePathEnd)
        return isRepeatable;
      if (isRepeatable)
        isRepeatable = !specNode.isSequence();
      specNode = specNode.getChild(childIndex);
      pathIndex++;
    }
  }

  // Adds to a set all the element names that can be used
  //   for replacing an element that occupy the position given
  //   by an element name path, relative to a pattern.
  // Each element name may be the first element name
  //   of a sequence of several required element names.
  public static final void addElementNamesForReplace(
    final Set elementNameSet, final SpecificationNode rootSpecNode,
    final int[] elementNamePath) {
    SpecificationNode specNode = rootSpecNode;
    int pathIndex = 0;
    while (true) {
      int childIndex = elementNamePath[pathIndex];
      boolean isElementNamePathEnd = childIndex < 0;
      if ((specNode.getChildCount() == 0) != isElementNamePathEnd)
        // The element name path end must map to a leaf spec node
        throw new IllegalArgumentException();
      if (isElementNamePathEnd)
        return;
      if (!specNode.isSequence())
        for (int k = 0; k < specNode.getChildCount(); k++)
          if (k != childIndex)
            addFirstElementNames(elementNameSet, specNode.getChild(k));
      specNode = specNode.getChild(childIndex);
      pathIndex++;
    }
  }

  // Adds to a set all the element names that can be used
  //   for replacing an element that repeats the pattern of an element
  //   that occupy the position given by an element name path,
  //   relative to a pattern.
  // Each element name may be the first element name
  //   of a sequence of several required element names.
  public static final void addElementNamesForReplaceRepeatedUse(
    final Set elementNameSet, final SpecificationNode rootSpecNode,
    final int[] elementNamePath) {
    SpecificationNode specNode = rootSpecNode;
    boolean isRepeatable = false;
    int pathIndex = 0;
    while (true) {
      int childIndex = elementNamePath[pathIndex];
      boolean isElementNamePathEnd = childIndex < 0;
      if ((specNode.getChildCount() == 0) != isElementNamePathEnd)
        // The element name path end must map to a leaf spec node
        throw new IllegalArgumentException();
      if (isElementNamePathEnd)
        return;
      if (specNode.isSequence())
        isRepeatable = false;
      else {
        if (!isRepeatable)
          isRepeatable = isOccurrenceRepeatable(specNode.getOccurrence());
        if (isRepeatable)
          for (int k = 0; k < specNode.getChildCount(); k++)
            if (k != childIndex)
              addFirstElementNames(elementNameSet, specNode.getChild(k));
      }
      specNode = specNode.getChild(childIndex);
      pathIndex++;
    }
  }

  // Get the element names of a SpecificationNode which may be inserted
  //   between two element names referenced by their element name paths.
  // Each element name may be the first element name
  //   of a sequence of several required element names.
  // Doesn't take into account the element names
  //   which can replace the two base element names.
  // Adds this element names to an element name set.
  public static final void addElementNamesInsertableBetween(
    final Set elementNameSet, final SpecificationNode rootSpecNode,
    final int[] elementNamePath1, final int[] elementNamePath2) {
    if (elementNamePath1 == null && elementNamePath2 == null) {
      addElementNames(elementNameSet, rootSpecNode);
      return;
    }
    if (elementNamePath1 == null) {
      addElementNamesInsertableBefore(elementNameSet, rootSpecNode,
        elementNamePath2, 0);
      return;
    }
    if (elementNamePath2 == null) {
      addElementNamesInsertableAfter(elementNameSet, rootSpecNode,
        elementNamePath1, 0);
      return;
    }
    SpecificationNode specNode = rootSpecNode;
    int pathIndex = 0;
    while (true) {
      int childIndex1 = elementNamePath1[pathIndex];
      int childIndex2 = elementNamePath2[pathIndex];
      boolean isLeaf = specNode.getChildCount() == 0;
      if (isLeaf || childIndex1 < 0 || childIndex2 < 0)
        if (isLeaf && childIndex1 < 0 && childIndex2 < 0
          && childIndex1 == childIndex2)
          return;
        else // The element name path end must map to a leaf spec node
          throw new IllegalArgumentException();
      if (specNode.isSequence()) {
        if (childIndex1 > childIndex2)
          throw new IllegalArgumentException();
        if (childIndex1 < childIndex2) {
          addElementNamesInsertableAfter(elementNameSet,
            specNode.getChild(childIndex1), elementNamePath1, pathIndex + 1);
          for (int k = childIndex1 + 1; k < childIndex2; k++)
            addFirstElementNames(elementNameSet, specNode.getChild(k));
          addElementNamesInsertableBefore(elementNameSet,
            specNode.getChild(childIndex2), elementNamePath2, pathIndex + 1);
          return;
        }
      }
      if (childIndex2 != childIndex1)
        return;
      specNode = specNode.getChild(childIndex1);
      pathIndex++;
    }
  }

  // Get the element names of a SpecificationNode which may be inserted
  //   before an element name referenced by its element name path.
  // Each element name may be the first element name
  //   of a sequence of several required element names.
  // Doesn't take into account the element names
  //   which can replace the base element name.
  // Adds this element names to an element name set.
  private static final void addElementNamesInsertableBefore(
    final Set elementNameSet, final SpecificationNode rootSpecNode,
    final int[] elementNamePath, final int startIndex) {
    SpecificationNode specNode = rootSpecNode;
    int pathIndex = startIndex;
    while (true) {
      int childIndex = elementNamePath[pathIndex];
      boolean isElementNamePathEnd = childIndex < 0;
      if ((specNode.getChildCount() == 0) != isElementNamePathEnd)
        // The element name path end must map to a leaf spec node
        throw new IllegalArgumentException();
      if (isElementNamePathEnd)
        return;
      if (specNode.isSequence())
        for (int k = 0; k < childIndex; k++)
          addFirstElementNames(elementNameSet, specNode.getChild(k));
      specNode = specNode.getChild(childIndex);
      pathIndex++;
    }
  }

  // Get the element names of a SpecificationNode which may be inserted
  //   after an element name referenced by its element name path.
  // Each element name may be the first element name
  //   of a sequence of several required element names.
  // Doesn't take into account the element names
  //   which can replace the base element name.
  // Adds this element names to an element name set.
  private static final void addElementNamesInsertableAfter(
    final Set elementNameSet, final SpecificationNode rootSpecNode,
    final int[] elementNamePath, final int startIndex) {
    SpecificationNode specNode = rootSpecNode;
    int pathIndex = startIndex;
    while (true) {
      int childIndex = elementNamePath[pathIndex];
      boolean isElementNamePathEnd = childIndex < 0;
      if ((specNode.getChildCount() == 0) != isElementNamePathEnd)
        // The element name path end must map to a leaf spec node
        throw new IllegalArgumentException();
      if (isElementNamePathEnd)
        return;
      if (specNode.isSequence())
        for (int k = childIndex + 1; k < specNode.getChildCount(); k++)
          addFirstElementNames(elementNameSet, specNode.getChild(k));
      specNode = specNode.getChild(childIndex);
      pathIndex++;
    }
  }
}
