/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.document.elementKey;

import java.util.Iterator;
import java.util.Set;
import java.util.Map;
import java.util.HashSet;
import java.util.HashMap;

import org.w3c.dom.Node;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Attr;
import org.w3c.dom.Element;

import org.xmloperator.framework.document.elementKey.ElementKey;
import org.xmloperator.document.elementKey.ElementKeyImpl;
import org.xmloperator.document.elementKey.ElementKeyUtil;

/**
 * An ElementKeyStat object keeps statictics on elements with a given name
 *   of a document in order to allow an automatic ElementKey selection.
 */
public final class ElementKeyStat {
  private int elementCount = 0;
  private boolean isLeafElement = true;
  private final Map fromAttributeNameToKeyStat = new HashMap();
  private final Map fromChildElementNameToKeyStat = new HashMap();

  public static final ElementKey getElementKey(
      Node parentNode, String elementName) {
    ElementKeyStat elementKeyStat = new ElementKeyStat();
    elementKeyStat.add(parentNode, elementName);
    ElementKey elementKey = elementKeyStat.getElementKey();
    elementKeyStat.clear();
    return elementKey;
  }

  /**
   * A key is an attribute or a single element child, identified by its name.
   */
  private static class KeyStat {
    private final String keyName;
    private int keyCount = 0;
    private final Set values = new HashSet();

    public KeyStat(String keyName) {
      if (keyName == null)
        throw new IllegalArgumentException();
      this.keyName = keyName;
    }

    public void clear() {
      this.keyCount = 0;
      this.values.clear();
    }

    public void addValue(String value) {
      if (value == null)
        throw new IllegalArgumentException();
      this.keyCount++;
      this.values.add(value);
    }

    public String getKeyName() {
      return this.keyName;
    }

    public int getKeyCount() {
      return this.keyCount;
    }

    public int getValueCount() {
      return this.values.size();
    }
  }

  /**
   * A multiple key is a multiple element child with the same name
   *   within the same element.
   */
  private static final KeyStat MULTIPLE_KEY_STAT = new KeyStat("");

  public ElementKeyStat() {
  }

  ////////////////////////
  // Making the statistics
  ////////////////////////

  public void clear() {
    this.elementCount = 0;
    this.isLeafElement = true;
    this.clearKeyStatMap(this.fromAttributeNameToKeyStat);
    this.clearKeyStatMap(this.fromChildElementNameToKeyStat);
  }

  private void clearKeyStatMap(Map keyStatMap) {
    Iterator iterator = keyStatMap.values().iterator();
    while (iterator.hasNext())
      ((KeyStat)iterator.next()).clear();
    keyStatMap.clear();
  }

  private static final int enoughElementCount = 100;

  public void add(Node parentNode, String elementName) {
    Node childNode = parentNode.getFirstChild();
    while (childNode != null && this.elementCount < enoughElementCount) {
      if (childNode.getNodeType() == Node.ELEMENT_NODE) {
        if (childNode.getNodeName().equals(elementName)) {
          this.elementCount++;
          this.addAttributes((Element)childNode);
          this.addChildElements((Element)childNode);
        }
        this.add(childNode, elementName);
      }
      childNode = childNode.getNextSibling();
    }
  }

  private void addAttributes(Element element) {
    NamedNodeMap attributes = element.getAttributes();
    for (int k = 0; k < attributes.getLength(); k++) {
      this.isLeafElement = false;
      Attr attr = (Attr)attributes.item(k);
      String attrValue = attr.getValue();
      if (attrValue.length() != 0) {
        String attrName = attr.getName();
        KeyStat keyStat =
            (KeyStat)this.fromAttributeNameToKeyStat.get(attrName);
        if (keyStat == null) {
          keyStat = new KeyStat(attrName);
          this.fromAttributeNameToKeyStat.put(attrName, keyStat);
        }
        keyStat.addValue(attrValue);
      }
    }
  }

  private void addChildElements(Element element) {
    Node childNode = element.getFirstChild();
    while (childNode != null) {
      switch (childNode.getNodeType()) {
        case Node.ELEMENT_NODE: {
          this.isLeafElement = false;
          if (!ElementKeyUtil.hasElementOrEntityRef( (Element) childNode)) {
            String name = childNode.getNodeName();
            if (ElementKeyUtil.getSingleElement(element, name) == null)
              this.fromChildElementNameToKeyStat.put(name, MULTIPLE_KEY_STAT);
            else {
              String value = ElementKeyUtil.leafElementText((Element)childNode);
              if (value.length() != 0) {
                KeyStat keyStat =
                    (KeyStat)this.fromChildElementNameToKeyStat.get(name);
                if (keyStat == null) {
                  keyStat = new KeyStat(name);
                  this.fromChildElementNameToKeyStat.put(name, keyStat);
                }
                keyStat.addValue(value);
              }
            }
          }
          break;
        }
        case Node.ENTITY_REFERENCE_NODE:
          this.isLeafElement = false;
          break;
      }
      childNode = childNode.getNextSibling();
    }
  }

  ///////////////////////
  // Using the statistics
  ///////////////////////

  private static final int minElementCount = 3;

  public ElementKey getElementKey() {
    if (this.isLeafElement)
      return new ElementKeyImpl(ElementKey.KEYTYPE_TEXTIFLEAF);
    if (this.elementCount < minElementCount)
      return null;
    KeyStat bestAttributeKeyStat =
        bestKeyStat(this.fromAttributeNameToKeyStat);
    KeyStat bestElementKeyStat =
        bestKeyStat(this.fromChildElementNameToKeyStat);
    ElementKey attributeElementKey = bestAttributeKeyStat != null
        && bestAttributeKeyStat.getKeyCount() == this.elementCount
        && 2 * bestAttributeKeyStat.getValueCount() >= this.elementCount ?
        new ElementKeyImpl(
        ElementKey.KEYTYPE_ONEATTRIBUTE, bestAttributeKeyStat.getKeyName()) :
        null;
    ElementKey elementElementKey = bestElementKeyStat != null
        && bestElementKeyStat.getKeyCount() == this.elementCount
        && 2 * bestElementKeyStat.getValueCount() >= this.elementCount ?
        new ElementKeyImpl(
        ElementKey.KEYTYPE_ONESINGLEELEMENT, bestElementKeyStat.getKeyName()) :
        null;
    if (attributeElementKey != null && elementElementKey != null)
      return bestAttributeKeyStat.getValueCount() >=
          bestElementKeyStat.getValueCount() ?
          attributeElementKey : elementElementKey;
    if (attributeElementKey != null)
      return attributeElementKey;
    if (elementElementKey != null)
      return elementElementKey;
    return null;
  }

  private static KeyStat bestKeyStat(Map keyStatMap) {
    KeyStat bestKeyStat = null;
    int bestValueCount = 0;
    String bestKeyName = "";
    Iterator iterator = keyStatMap.values().iterator();
    while (iterator.hasNext()) {
      KeyStat keyStat = (KeyStat) iterator.next();
      String keyName = keyStat.getKeyName();
      int valueCount = keyStat.getValueCount();
      if (valueCount > bestValueCount
          || (valueCount == bestValueCount
              && keyStat != MULTIPLE_KEY_STAT
              && keyName.compareTo(bestKeyName) > 0)) {
        bestKeyStat = keyStat;
        bestValueCount = valueCount;
        bestKeyName = keyName;
      }
    }
    return bestKeyStat;
  }
}
