/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.document.elementKey;

import org.w3c.dom.Node;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Attr;
import org.w3c.dom.Element;
import org.w3c.dom.DocumentType;

import org.xmloperator.framework.document.elementKey.ElementKey;

public abstract class ElementKeyUtil {

  /**
   * Return a string that represents the given <code>node</code>.
   */
  public static final String nodeToString(Node node, ElementKey elementKey) {
    if (node == null)
      return "";
    switch (node.getNodeType()) {
      case Node.DOCUMENT_NODE:
        return "DOCUMENT";
      case Node.DOCUMENT_TYPE_NODE:
        return documentTypeNodeToString((DocumentType)node);
      case Node.ELEMENT_NODE:
        return elementToString((Element)node, elementKey);
      case Node.TEXT_NODE:
        return node.getNodeValue().trim();
      case Node.CDATA_SECTION_NODE:
        return "<![CDATA[" + node.getNodeValue() + "]]>";
      case Node.COMMENT_NODE:
        return "<!--" + node.getNodeValue() + "-->";
      case Node.ENTITY_REFERENCE_NODE:
        return "&" + node.getNodeName() + ";";
      case Node.PROCESSING_INSTRUCTION_NODE:
        return "<?" + node.getNodeName() + " " + node.getNodeValue() + "?>";
      default:
        return "";
    }
  }

  private static final String documentTypeNodeToString(
      DocumentType documentType) {
    StringBuffer stringBuffer = new StringBuffer();
    stringBuffer.append("DOCTYPE");
    String publicId = documentType.getPublicId();
    String systemId = documentType.getSystemId();
    if (publicId != null && publicId.length() != 0) {
      stringBuffer.append(" publicId=\"");
      stringBuffer.append(publicId);
      stringBuffer.append('"');
    }
    if (systemId != null && systemId.length() != 0) {
      stringBuffer.append(" systemId=\"");
      stringBuffer.append(systemId);
      stringBuffer.append('"');
    }
    return stringBuffer.toString();
  }

  private static final String elementToString(
      Element element, ElementKey elementKey) {
    StringBuffer sb = new StringBuffer();
    sb.append("< ");
    sb.append(element.getNodeName());
    if (elementKey == null)
      addAllAttributes(sb, element);
    else
      switch (elementKey.getKeyType()) {
        case ElementKey.KEYTYPE_TEXTIFLEAF:
          addValue(sb, leafElementText(element));
          break;
        case ElementKey.KEYTYPE_ALLATTRIBUTES:
          addAllAttributes(sb, element);
          break;
        case ElementKey.KEYTYPE_ALLSINGLEELEMENTS:
          addAllSingleElements(sb, element);
          break;
        case ElementKey.KEYTYPE_ALLATTRIBUTESANDSINGLEELEMENTS:
          addAllAttributes(sb, element);
          addAllSingleElements(sb, element);
        case ElementKey.KEYTYPE_ONEATTRIBUTE: {
          String name = elementKey.getAttributeOrSingleElementName();
          String value = element.getAttribute(name);
          if (value != null)
            addAttributeOrSingleElement(sb, true, name, value);
          break;
        }
        case ElementKey.KEYTYPE_ONESINGLEELEMENT: {
          String name = elementKey.getAttributeOrSingleElementName();
          Element singleElement = getSingleElement(element, name);
          if (singleElement != null)
            addAttributeOrSingleElement(
                sb, false, name, leafElementText(singleElement));
          break;
        }
        default:
          throw new IllegalArgumentException();
      }
    sb.append(" >");
    return sb.toString();
  }

  public static final String leafElementText(Element element) {
    Node firstChild = element.getFirstChild();
    if (firstChild == null)
      return "";
    if (firstChild.getNodeType() == Node.TEXT_NODE
        && firstChild.getNextSibling() == null)
      return firstChild.getNodeValue();
    if (hasElementOrEntityRef(element))
      return "";
    StringBuffer sb = new StringBuffer();
    Node childNode = firstChild;
    while (childNode != null) {
      switch (childNode.getNodeType()) {
        case Node.CDATA_SECTION_NODE:
        case Node.TEXT_NODE:
          sb.append(childNode.getNodeValue());
          break;
      }
      childNode = childNode.getNextSibling();
    }
    return sb.toString().trim();
  }

  public static final boolean hasElementOrEntityRef(Element element) {
    Node childNode = element.getFirstChild();
    while (childNode != null) {
      int nodeType = childNode.getNodeType();
      if (nodeType == Node.ELEMENT_NODE
          || nodeType == Node.ENTITY_REFERENCE_NODE)
        return true;
      childNode = childNode.getNextSibling();
    }
    return false;
  }

  private static final void addAllAttributes(StringBuffer sb, Element element) {
    NamedNodeMap attrs = element.getAttributes();
    for (int k = 0; k < attrs.getLength(); k++) {
      Attr attr = (Attr) attrs.item(k);
      addAttributeOrSingleElement(sb, true, attr.getName(), attr.getValue());
    }
  }

  private static final void addAllSingleElements(
      StringBuffer sb, Element element) {
    Node childNode = element.getFirstChild();
    while (childNode != null) {
      if (childNode.getNodeType() == Node.ELEMENT_NODE) {
        String childElementName = childNode.getNodeName();
        if (getSingleElement(element, childElementName) != null)
          addAttributeOrSingleElement(
              sb, false, childElementName, leafElementText((Element)childNode));
      }
      childNode = childNode.getNextSibling();
    }
  }

  public static final Element getSingleElement(
      Node parentNode, String childElementName) {
    Element childElement = null;
    Node childNode = parentNode.getFirstChild();
    while (childNode != null) {
      if (childNode.getNodeType() == Node.ELEMENT_NODE
          && childNode.getNodeName().equals(childElementName))
        if (childElement != null)
          return null;
        else
          childElement = (Element)childNode;
      childNode = childNode.getNextSibling();
    }
    return childElement;
  }

  private static final void addValue(StringBuffer sb, String value) {
    if (value.length() == 0)
      return;
    sb.append("=\"");
    sb.append(value);
    sb.append('"');
  }

  private static final void addAttributeOrSingleElement(
      StringBuffer sb, boolean isAttribute, String name, String value) {
    if (value.length() == 0)
      return;
    sb.append(' ');
    if (isAttribute)
      sb.append('@');
    sb.append(name);
    addValue(sb, value);
  }
}
