/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.document.manage;

import java.util.Map;
import java.util.HashMap;

import org.w3c.dom.Node;
import org.w3c.dom.Element;

import org.xmloperator.framework.document.namespace.UndefinedNamespaceException;
import org.xmloperator.framework.document.namespace.QualifiedName;
import org.xmloperator.framework.document.relaxng.NameClass;
import org.xmloperator.framework.document.relaxng.Pattern;
import org.xmloperator.framework.document.relaxng.NonEmptyPattern;
import org.xmloperator.framework.document.relaxng.Ref;
import org.xmloperator.framework.document.relaxng.GroupPattern;
import org.xmloperator.framework.document.relaxng.ElementDefinition;
import org.xmloperator.framework.document.relaxng.Schema;
import org.xmloperator.framework.document.id.IdException;
import org.xmloperator.framework.document.id.IdManager;
import org.xmloperator.framework.document.schema.ElementContent;
import org.xmloperator.framework.document.schema.AttributesContent;
import org.xmloperator.framework.document.schema.
    NotConformantToContentModelException;
import org.xmloperator.framework.document.schema.
    InvalidElementSpecificationException;
import org.xmloperator.framework.document.manage.ElementDefinitionMapper;
import org.xmloperator.framework.document.manage.AttributeChangeListener;
import org.xmloperator.framework.document.manage.AttributesContentAccess;
import org.xmloperator.util.dom.ElementIndexer;
import org.xmloperator.util.dom.QualifiedNameImpl;
import org.xmloperator.util.dom.ElementBrowser;
import org.xmloperator.document.relaxng.NameClassUtils;
import org.xmloperator.document.relaxng.ElementContentImpl;
import org.xmloperator.document.relaxng.AttributesContentImpl;
import org.xmloperator.document.manage.AbstractElementManager;
import org.xmloperator.document.manage.EmptyElementContent;

public final class RNGElementManager extends AbstractElementManager {
  private Schema schema;
  private final Map fromElementToElementDefinition = new HashMap();

  public RNGElementManager(IdManager idManager, Schema schema) {
    super(idManager);
    if (schema == null)
      throw new IllegalArgumentException();
    this.schema = schema;
  }

  protected void addElementContentToCache(Element parentElement)
    throws UndefinedNamespaceException, NotConformantToContentModelException {
    if (contentCache.getElementContent(parentElement) != null)
      return;
    ElementDefinition parentElementDefinition =
      (ElementDefinition)getElementDefinition(parentElement);
    if (parentElementDefinition == null)
      return;
    Pattern pattern = parentElementDefinition.getPattern();
    if (pattern == null)
      return;
    boolean isEmptyPattern = pattern.isEmpty();
    if (isEmptyPattern && ElementBrowser.hasChildElements(parentElement))
      throw new InvalidElementSpecificationException(0);
    contentCache.put(parentElement,
      isEmptyPattern ?
        (ElementContent)EmptyElementContent.EMPTY_CHILDREN_CONTENT :
        (ElementContent)new ElementContentImpl((NonEmptyPattern)pattern,
        QualifiedNameImpl.qualifiedNames(parentElement),
        QualifiedNameImpl.attributeQualifiedNames(parentElement)));
  }

  protected void addAttributesContentToCache(Element parentElement) {
    if (contentCache.getAttributesContent(parentElement) != null)
      return;
    ElementDefinition parentElementDefinition =
      (ElementDefinition)getElementDefinition(parentElement);
    if (parentElementDefinition == null)
      return;
    Pattern pattern = parentElementDefinition.getPattern();
    if (pattern == null || pattern.isEmpty())
      return;
    contentCache.put(
        parentElement,
        new AttributesContentImpl(parentElement, (NonEmptyPattern)pattern));
  }

  protected void removeElementDefinitionFromCache(Element element) {
    fromElementToElementDefinition.remove(element);
  }

  /////////////////////////////////////////
  // ElementDefinitionMapper implementation
  /////////////////////////////////////////

  public Object getSchema() {
    return schema;
  }

  public Object getElementDefinition(Element element) {
    ElementDefinition elementDefinition =
      (ElementDefinition)fromElementToElementDefinition.get(element);
    if (elementDefinition != null)
      return elementDefinition;
    Node parentNode = element.getParentNode();
    if (parentNode == null)
      return null;
    int parentNodeType = parentNode.getNodeType();
    if (parentNodeType == Node.DOCUMENT_NODE) {
      Pattern start = schema.getStart();
      if (start == null || start.isEmpty())
        return null;
      try {
        elementDefinition = startingElementDefinition(
          new QualifiedNameImpl(element), (NonEmptyPattern)start);
      }
      catch (UndefinedNamespaceException exception) {
        return null;
      }
      if (elementDefinition != null)
        fromElementToElementDefinition.put(element, elementDefinition);
      return elementDefinition;
    }
    if (parentNodeType != Node.ELEMENT_NODE)
      return null;
    ElementContent elementContent = null;
    try {
      elementContent = getElementContent((Element)parentNode);
      if (elementContent == null)
        return null;
    }
    catch (NotConformantToContentModelException exception) {
      return null;
    }
    Node childNode = parentNode.getFirstChild();
    while (childNode != null) {
      if (childNode.getNodeType() == Node.ELEMENT_NODE) {
        int flatIndex = ElementIndexer.getFlatIndex(childNode);
        ElementDefinition childDefinition =
          (ElementDefinition)elementContent.getElementDefinition(flatIndex);
        fromElementToElementDefinition.put(childNode, childDefinition);
        if (element == childNode)
          elementDefinition = childDefinition;
      }
      childNode = childNode.getNextSibling();
    }
    return elementDefinition;
  }

  private static final ElementDefinition startingElementDefinition(
    QualifiedName qualifiedName, NonEmptyPattern pattern) {
    switch (pattern.getPatternType()) {
      case NonEmptyPattern.PATTERN_REF: {
        ElementDefinition elementDefinition =
          ((Ref)pattern).getElementDefinition();
        return NameClassUtils.isNameContained(elementDefinition.getNameClass(),
          qualifiedName.getNamespaceURI(), qualifiedName.getLocalName()) ?
          elementDefinition : null;
      }
      case NonEmptyPattern.PATTERN_GROUP: {
        int groupType = ((GroupPattern)pattern).getGroupType();
        if (groupType != GroupPattern.GROUP_CHOICE_EXCLUDING_EMPTY_PATTERN
          && groupType != GroupPattern.GROUP_CHOICE_INCLUDING_EMPTY_PATTERN)
          return null;
        for (int k = 0; k < ((GroupPattern)pattern).getChildCount(); k++) {
          ElementDefinition elementDefinition = startingElementDefinition(
            qualifiedName, ((GroupPattern)pattern).getChild(k));
          if (elementDefinition != null)
            return elementDefinition;
        }
        return null;
      }
      default:
        return null;
    }
  }
}
