/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.framework.control.instance;

import org.w3c.dom.Node;
import org.w3c.dom.Element;
import org.w3c.dom.Text;

import org.xmloperator.framework.document.namespace.NamespacePrefixPair;

public interface EditControl {
  // Insertion types
  public static final int INSERTION_PREVIOUS = 0;
  public static final int INSERTION_NEXT = 1;
  public static final int INSERTION_CHILD = 2;
  // Move types
  public static final int MOVE_PREVIOUS = 0;
  public static final int MOVE_NEXT = 1;
  public static final int MOVE_CHILD = 2;
  public static final int MOVE_PARENT = 3;

  /**
   * Set the readonly mode.
   */
  public void setReadOnly(boolean isReadOnly);

  /**
   * Set the lazy mode.
   */
  public void setLazy(boolean isLazy);

  /**
   * Set the XML version.
   */
  public void setXMLVersion(String xMLVersion);

  /**
   * Set the document encoding.
   */
  public void setEncoding(String encoding);

  /**
   * Set the dosument standalone attribute.
   */
  public void setStandalone(boolean isStandalone);

  /**
   * Insert a new node upon insertionType (== INSERTION_*),
   *   relatively to relativeNode.
   * If relativeNode is null then uses the selected node (if any) in place of it.
   * The operation is reversible.
   */
  public void insertNewNode(Node relativeNode, int insertionType,
    String elementName);

  /**
   * Delete the given nodes.
   * If nodes is null then uses the selected nodes (if any) in place of them.
   * The operation is reversible.
   */
  public void deleteNodeInterval(Node[] nodes);

  /**
   * Go to move mode, using the selected nodes.
   */
  public void storeSelectedNodesForMove();

  /**
   * Quit the move mode.
   */
  public void clearNodeIntervalForMove();

  /**
   * Move the given nodes accordingly to moveType (== MOVE_*).
   * If nodes is null then uses the selected nodes (if any) in place of them.
   * The operation is reversible.
   */
  public void moveNodeInterval(Node[] nodes, int moveType);

  /**
   * Reorder the given nodes accordingly to the value of the attribute
   *   whose name list is given.
   * If nodes is null then uses the selected nodes (if any) in place of them.
   * The operation is reversible.
   */
  public void reorderNodeInterval(Node[] nodes, String attributeNames,
    boolean isReverse);

  /**
   * Pack the given nodes within a new element replacing them.
   * If nodes is null then uses the selected nodes (if any) in place of them.
   * The operation is reversible.
   */
  public void packNodeInterval(Node[] nodes, String elementName);

  /**
   * Replace the element by its children (if any).
   * If element is null then uses the selected element (if any) in place of it.
   * The operation is reversible.
   */
  public void unpackElement(Element element);

  /**
   * Remove the given nodes and put them in the document clipboard.
   * Copy also to system clipboard, on demand.
   * If nodes is null then uses the selected nodes (if any) in place of them.
   * The operation is reversible.
   */
  public void cutNodes(Node[] nodes, boolean copyToSystemClipoard);

  /**
   * Copy the given nodes in the document clipboard.
   * Copy also to system clipboard, on demand.
   * If nodes is null then uses the selected nodes (if any) in place of them.
   */
  public void copyNodes(Node[] nodes, boolean copyToSystemClipoard);

  /**
   * Insert the document clipboard content upon insertionType,
   *   relatively to relativeNode.
   * Make first a copy from system clipboard, on demand.
   * If relativeNode is null then uses the selected node (if any) in place of it.
   * The operation is reversible.
   */
  public void pasteDocumentFragment(Node relativeNode,
    int insertionType, boolean copyFromSystemClipboard);

  /**
   * Replace the given nodes by the document fragment.
   * Make first a copy from system clipboard, on demand.
   * If nodes is null then replace the selected nodes (if any) in place of them.
   * The operation is reversible.
   */
  public void replaceNodesByDocumentFragment(Node[] nodes,
      boolean copyFromSystemClipboard);

  /**
   * Clear the document clipboard.
   */
  public void clearDocumentFragment();

  /**
   * Serialize and copy the document clipboard to the system clipboard.
   * The entity references, if any, are expanded.
   * Namespace declarations may missing.
   */
  public void copyDocumentFragmentToSystemClipboard();

  /**
   * Parse and copy the system clipboard to the document cliboard.
   * The document clipboard is cleared if there is no well formed XML
   *   in the system clipboard.
   */
  public void copySystemClipboardToDocumentFragment();

  /**
   * Set node to a new value.
   * node and newValue must be not null.
   * The operation is reversible.
   */
  public void setNodeValue(Node node, String newValue);

  /**
   * Create a new attribute node for element.
   * If element is null then uses the selected element (if any) in place of it.
   * If attributeName is null then ask for it.
   * The operation is reversible.
   */
  public void addNewAttribute(Element element, String attributeName);

  /**
   * Set the value of the attribute of element whose name is attributeName
   *   to the newAttributeValue value.
   * If newAttributeValue is null then the attribute node is removed.
   * If element is null then uses the selected element (if any) in place of it.
   * If attributeName is null then uses the selected attribute name (if any)
   *   in place of it.
   * The operation is reversible.
   */
  public void setAttributeValue(Element element, String attributeName,
    String newAttributeValue);

  /**
   * Add a namespace declaration to the root element.
   * The operation is reversible.
   */
  public void addNamespace(NamespacePrefixPair namespacePrefixPair);

  /**
   * Replace the given node by a node of type given by nodeType
   *   and name (if any) given by nodeName.
   * If node is null then uses the selected node (if any) in place of it.
   * Keep childs in case of element to element replacement.
   * The operation is reversible.
   */
  public void replaceNode(Node node, int nodeType, String nodeName);

  /**
   * Join to sibling texts.
   * If text is null then uses the selected text node (if any) in place of it.
   * The operation is reversible.
   */
  public void joinText(Text text, boolean withNextSibling);

  /**
   * Split the given text at the offset position.
   * If text is null then uses the selected text node (if any) in place of it.
   * The operation is reversible.
   */
  public void splitText(Text text, int offset);

  /**
   * Execute the operator given by its name to the given element.
   * If element is null then uses the selected element (if any) in place of it.
   * The operations are reversible.
   */
  public void operateElement(Element element, String operatorName);

  /**
   * Execute the operator given by its index to the given element.
   * If element is null then uses the selected element (if any) in place of it.
   * The operations are reversible.
   */
  public void operateElement(Element element, int operatorIndex);

  /**
   * Add children to element in order to make it valid accordingly
   *   to the current schema.
   * If element is null then uses the selected element (if any) in place of it.
   * The operations are reversible.
   */
  public void addChildsIfNecessary(Element element);

  /**
   * Undo the operation on the top of the undoable operation stack.
   */
  public void undoOperation();

  /**
   * Undo all the operations from the undoable operation stack.
   */
  public void undoAllOperations();

  /**
   * Redo the operation on the top of the redoable operation stack.
   */
  public void redoOperation();

  /**
   * Redo all the operations from the redoable operation stack.
   */
  public void redoAllOperations();

  /**
   * Clear the undoable operation stack.
   */
  public void clearUndoOperations();

  /**
   * Clear the redoable operation stack.
   */
  public void clearRedoOperations();

  /**
   * Load an operation stack from a file.
   */
  public void loadOperationStack(String pathName);

  /**
   * Save an operation stack to a file.
   */
  public void saveOperationStack(String pathName);
}
