/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.operation;

import java.util.Vector;

import org.w3c.dom.Element;
import org.w3c.dom.Document;
import org.apache.xerces.dom.DocumentImpl;

import org.xmloperator.framework.document.namespace.UndefinedNamespaceException;
import org.xmloperator.framework.document.id.IdException;
import org.xmloperator.framework.operation.OperationException;
import org.xmloperator.framework.operation.Operation;
import org.xmloperator.framework.util.Observer;
import org.xmloperator.operation.core.TransitionOperation;
import org.xmloperator.operation.OperationPrinter;
import org.xmloperator.operation.OperationStack;

public final class OperationStackImpl implements OperationStack {
  private static final String NO_OPERATION = "No operation";
  private Observer observer = null;
  private OperationPrinter operationPrinter = null;
  private Vector operationList;
  private int redoCount;
  private Document operationDocument;

  public OperationStackImpl() {
    operationList = new Vector();
    redoCount = 0;
    operationDocument = new DocumentImpl();
  }

  ////////////////////////////////////////
  // OperationStackReadOnly implementation
  ////////////////////////////////////////

  public int getPossibleUndoCount() {
    return operationList.size() - redoCount;
  }

  public int getPossibleRedoCount() {
    return redoCount;
  }

  public String getUndoOperationName(int index) {
    int undoCount = operationList.size() - redoCount;
    if (index < 0 || index >= undoCount)
      return null;
    return ((Operation)operationList.elementAt(undoCount - 1 - index)).
      getOperationName();
  }

  public String getRedoOperationName(int index) {
    if (index < 0 || index >= redoCount)
      return null;
    int undoCount = operationList.size() - redoCount;
    return ((Operation)operationList.elementAt(undoCount + index)).
      getReverseOperationName();
  }

  ////////////////////////////////
  // OperationStack implementation
  ////////////////////////////////

  public void setObserver(Observer observer) {
    this.observer = observer;
  }

  public void setOperationPrinter(OperationPrinter operationPrinter) {
    this.operationPrinter = operationPrinter;
  }

  public void executeAndAdd(Operation operation)
    throws IdException, UndefinedNamespaceException {
    if (operation == null || operation.isExecuted())
      throw new IllegalArgumentException();
    if (operation.isNull())
      return;
    executeOperation(operation);
    clearRedoOperations();
    boolean isLastOperationHiden = false;
    int count = operationList.size();
    if (count != 0) {
      Operation lastOperation = (Operation)operationList.lastElement();
      isLastOperationHiden = operation.hidesIfCan(lastOperation);
    }
    if (isLastOperationHiden)
      operationList.setElementAt(operation, count - 1);
    else
      operationList.add(operation);
    if (observer != null) {
      int notification = isLastOperationHiden ?
        OperationStackReadOnly.NOTIFY_REPLACE :
        OperationStackReadOnly.NOTIFY_ADD;
      observer.update(this, new Integer(notification));
    }
  }

  public void clearUndoOperations() {
    while (operationList.size() > redoCount) {
      Operation operation = (Operation)operationList.elementAt(0);
      operationList.remove(0);
    }
    if (observer != null)
      observer.update(this,
        new Integer(OperationStackReadOnly.NOTIFY_CLEAR_UNDO));
  }

  public void clearRedoOperations() {
    while (redoCount > 0) {
      int lastIndex = operationList.size() - 1;
      Operation operation = (Operation)operationList.elementAt(lastIndex);
      operationList.remove(lastIndex);
      redoCount--;
    }
    if (observer != null)
      observer.update(this,
        new Integer(OperationStackReadOnly.NOTIFY_CLEAR_REDO));
  }

  public void clear() {
    clearUndoOperations();
    clearRedoOperations();
  }

  public void undo(boolean skipsTransitions)
    throws IdException, UndefinedNamespaceException {
    int undoCount = operationList.size() - redoCount;
    if (undoCount == 0)
      throw new OperationException(NO_OPERATION);
    while (true) {
      undoOneOperationWithNotification();
      undoCount--;
      if (undoCount == 0 || !(operationList.elementAt(undoCount - 1) instanceof
        TransitionOperation))
        break;
      undoOneOperationWithNotification();
      undoCount--;
      if (undoCount == 0 || !skipsTransitions)
        break;
    }
  }

  private void undoOneOperationWithNotification()
    throws IdException, UndefinedNamespaceException {
    undoOneOperation();
    if (observer != null)
      observer.update(this,
        new Integer(OperationStackReadOnly.NOTIFY_UNDO));
  }

  private void undoOneOperation()
    throws IdException, UndefinedNamespaceException {
    int undoCount = operationList.size() - redoCount;
    int index = undoCount - 1;
    Operation operation = (Operation)operationList.elementAt(index);
    Operation reverseOperation = operation.createReverseOperation();
    executeOperation(reverseOperation);
    operationList.setElementAt(reverseOperation, index);
    redoCount++;
  }

  public void undoAllOperations()
    throws IdException, UndefinedNamespaceException {
    OperationException exception = null;
    try {
      while (true) {
        int undoCount = operationList.size() - redoCount;
        if (undoCount == 0)
          break;
        undoOneOperation();
      }
    }
    catch (OperationException operationException) {
      exception = operationException;
    }
    if (observer != null)
      observer.update(this,
        new Integer(OperationStackReadOnly.NOTIFY_CLEAR_NEW));
    if (exception != null)
      throw exception;
  }

  public void redo(boolean skipsTransitions)
    throws IdException, UndefinedNamespaceException {
    redo(skipsTransitions, false);
  }

  public Element redo(boolean skipsTransitions,
    boolean isOperationElementReturned)
    throws IdException, UndefinedNamespaceException {
    if (redoCount == 0)
      throw new OperationException(NO_OPERATION);
    while (true) {
      Element operationElement =
        redoOneOperationWithNotification(isOperationElementReturned);
      if (isOperationElementReturned || redoCount == 0
        || !(operationList.elementAt(operationList.size() - redoCount)
          instanceof TransitionOperation))
        return operationElement;
      redoOneOperationWithNotification(false);
      if (redoCount == 0 || !skipsTransitions)
        return null;
    }
  }

  private Element redoOneOperationWithNotification(
    boolean isOperationElementReturned)
    throws IdException, UndefinedNamespaceException {
    Element operationElement = redoOneOperation(isOperationElementReturned);
    if (observer != null)
      observer.update(this,
        new Integer(OperationStackReadOnly.NOTIFY_REDO));
    return operationElement;
  }

  private Element redoOneOperation(boolean isOperationElementReturned)
    throws IdException, UndefinedNamespaceException {
    int index = operationList.size() - redoCount;
    Operation reverseOperation = (Operation)operationList.elementAt(index);
    Operation operation = reverseOperation.createReverseOperation();
    Element operationElement =
      executeOperation(operation, isOperationElementReturned);
    operationList.setElementAt(operation, index);
    redoCount--;
    return operationElement;
  }

  public void redoAllOperations()
    throws IdException, UndefinedNamespaceException {
    OperationException exception = null;
    try {
      while (redoCount > 0)
        redoOneOperation(false);
    }
    catch (OperationException operationException) {
      exception = operationException;
    }
    if (observer != null)
      observer.update(this,
        new Integer(OperationStackReadOnly.NOTIFY_CLEAR_NEW));
    if (exception != null)
      throw exception;
  }

  private void executeOperation(Operation operation)
    throws IdException, UndefinedNamespaceException {
    executeOperation(operation, false);
  }

  private Element executeOperation(Operation operation,
    boolean isOperationElementMustBeReturned)
    throws IdException, UndefinedNamespaceException {
    if (operationPrinter != null && operationPrinter.isReady())
      isOperationElementMustBeReturned = true;
    if (isOperationElementMustBeReturned) {
      Element operationElement = operation.execute(operationDocument);
      operationElement =
        createOperationElement(operationDocument, operationElement);
      if (operationPrinter != null)
        operationPrinter.printOperation(operation, operationElement);
      return operationElement;
    }
    else {
      operation.execute(null);
      return null;
    }
  }

  private static final Element createOperationElement(
    Document operationDocument, Element internalOperationElement) {
    Element operationElement = operationDocument.createElement("operation");
    operationElement.appendChild(internalOperationElement);
    return operationElement;
  }
}
