/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.util.dom;

import org.w3c.dom.Node;
import org.w3c.dom.Element;

public abstract class ElementIndexer {

  //////////////////////////////////////////////////////////////////////
  // "Significant node" means
  //   "Element or Entity reference that contains at least one element"
  // "Significant node count" means
  //   "Element count, including elements inside entity references"
  //////////////////////////////////////////////////////////////////////

  // Return the significant child count of parentNode.
  public static final int significantChildCount(Node parentNode) {
    int count = 0;
    Node childNode = parentNode.getFirstChild();
    while (childNode != null) {
      switch (childNode.getNodeType()) {
        case Node.ELEMENT_NODE:
          count++;
          break;
        case Node.ENTITY_REFERENCE_NODE:
          count+= significantChildCount(childNode);
          break;
      }
      childNode = childNode.getNextSibling();
    }
    return count;
  }

  // Return true if the parentNode has at least one significant child.
  public static final boolean hasSignificantChilds(Node parentNode) {
    Node childNode = parentNode.getFirstChild();
    while (childNode != null) {
      switch (childNode.getNodeType()) {
        case Node.ELEMENT_NODE:
          return true;
        case Node.ENTITY_REFERENCE_NODE:
          if (hasSignificantChilds(childNode))
            return true;
          break;
      }
      childNode = childNode.getNextSibling();
    }
    return false;
  }

  public static class NodeInsideEntityException extends Exception {
  }

  // Return the significant child at the flatChildIndex position.
  public static final Element getSignificantChild(Node parentNode,
    int flatChildIndex) throws NodeInsideEntityException {
    if (flatChildIndex < 0)
      throw new IllegalArgumentException();
    Node childNode = parentNode.getFirstChild();
    while (childNode != null) {
      int significantCount = significantCount(childNode);
      if (significantCount != 0) {
        if (flatChildIndex == 0 && childNode.getNodeType() == Node.ELEMENT_NODE)
          return (Element)childNode;
        flatChildIndex-= significantCount;
        if (flatChildIndex < 0)
          throw new NodeInsideEntityException();
      }
      childNode = childNode.getNextSibling();
    }
    return null;
  }

  // Search for the first significant node from node,
  //   then return the position of this significant node.
  public static final int getFlatIndex(Node node) {
    if (node == null)
      return -1;
    int index = 0;
    while (true) {
      node = node.getPreviousSibling();
      if (node == null)
        return index;
      index+= significantCount(node);
    }
  }

  // Return the count of elements that represents "node".
  public static final int significantCount(Node node) {
    switch (node.getNodeType()) {
      case Node.ELEMENT_NODE:
        return 1;
      case Node.ENTITY_REFERENCE_NODE:
        return significantChildCount(node);
      default:
        return 0;
    }
  }

  // return true if "node" represents at least one element.
  public static final boolean isSignificant(Node node) {
    switch (node.getNodeType()) {
      case Node.ELEMENT_NODE:
        return true;
      case Node.ENTITY_REFERENCE_NODE:
        return hasSignificantChilds(node);
      default:
        return false;
    }
  }

  // Return the first significant node from node.
  public static final Node getNextSignificantSibling(Node node) {
    while (node != null) {
      if (isSignificant(node))
        return node;
      node = node.getNextSibling();
    }
    return null;
  }

  // Return the first significant node from node, in reverse order.
  public static final Node getPreviousSignificantSibling(Node node) {
    while (node != null) {
      if (isSignificant(node))
        return node;
      node = node.getPreviousSibling();
    }
    return null;
  }

  // return true if a node at least is significant.
  public static final boolean hasSignificantNodes(Node[] nodes) {
    for (int k = 0; k < nodes.length; k++)
      if (isSignificant(nodes[k]))
        return true;
    return false;
  }

  // Return the first significant node.
  public static final Node getFirstSignificantNode(Node[] nodes) {
    for (int k = 0; k < nodes.length; k++) {
      Node node = nodes[k];
      if (isSignificant(node))
        return node;
    }
    return null;
  }

  // Return the last significant node.
  public static final Node getLastSignificantNode(Node[] nodes) {
    for (int k = nodes.length - 1; k >= 0; k--) {
      Node node = nodes[k];
      if (isSignificant(node))
        return node;
    }
    return null;
  }

  // Return the position of the last significant node.
  public static final int getLastFlatIndex(Node[] nodes) {
    return getFlatIndex(getLastSignificantNode(nodes));
  }
}
