/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.util.dom;

import java.util.Arrays;
import java.util.Set;
import java.util.HashSet;

import org.w3c.dom.Node;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Attr;
import org.w3c.dom.Element;

import org.xmloperator.framework.document.namespace.UndefinedNamespaceException;

public abstract class NamespaceUtils {

  // Return the prefix (may be empty) of the qualified name.
  public static final String prefixFromQName(final String qName) {
    int index = qName.indexOf(':');
    return index >= 0 ? qName.substring(0, index) : "";
  }

  // Return the local name of the qualified name.
  public static final String localNameFromQName(final String qName) {
    int index = qName.indexOf(':');
    return index >= 0 ? qName.substring(index + 1) : qName;
  }

  // Return the default namespace URI (may be empty) in the context of element.
  public static final String defaultNamespace(Element element) {
    if (element == null)
      throw new IllegalArgumentException();
    while (true) {
      {
        Attr attributeNode = element.getAttributeNode("xmlns");
        if (attributeNode != null)
          return attributeNode.getNodeValue();
      }
      {
        Node parentNode = element.getParentNode();
        if (parentNode == null || parentNode.getNodeType() != Node.ELEMENT_NODE)
          return "";
        element = (Element)parentNode;
      }
    }
  }

  // Return the namespace URI (may be empty) which maps to the prefix
  //   in the context of element.
  // Throws a UndefinedNamespaceException if undefined.
  public static final String prefixToNamespace(Element element,
    final String prefix) throws UndefinedNamespaceException {
    if (element == null || prefix == null || prefix.length() == 0)
      throw new IllegalArgumentException();
    if (prefix.equals("xmlns"))
      return "http://www.w3.org/2000/xmlns/";
    if (prefix.equals("xml"))
      return "http://www.w3.org/XML/1998/namespace";
    String attributeName = "xmlns:" + prefix;
    while (true) {
      {
        Attr attributeNode = element.getAttributeNode(attributeName);
        if (attributeNode != null)
          return attributeNode.getNodeValue();
      }
      {
        Node parentNode = element.getParentNode();
        if (parentNode == null || parentNode.getNodeType() != Node.ELEMENT_NODE)
          throw new UndefinedNamespaceException(element, prefix);
        element = (Element)parentNode;
      }
    }
  }

  // Return the namespace URI (may be empty) of the attribute of element,
  //   with the given qualified name.
  // Throws a UndefinedNamespaceException if undefined.
  public static final String attrQNameToNamespace(final Element element,
    final String attrQName) throws UndefinedNamespaceException {
    if (attrQName.equals("xmlns"))
      return "http://www.w3.org/2000/xmlns/";
    String attrPrefix = prefixFromQName(attrQName);
    return attrPrefix.length() != 0 ?
      prefixToNamespace(element, attrPrefix) : "";
  }

  // Return the prefix that maps to the namespace in the context of element.
  // May be null if undefined.
  public static final String namespaceToPrefix(Element element,
    final String namespaceURI) {
    if (element == null || namespaceURI == null)
      throw new IllegalArgumentException();
    if (namespaceURI.equals("http://www.w3.org/2000/xmlns/"))
      return "xmlns";
    if (namespaceURI.equals("http://www.w3.org/XML/1998/namespace"))
      return "xml";
    while (true) {
      {
        NamedNodeMap attributes = element.getAttributes();
        for (int k = 0; k < attributes.getLength(); k++) {
          Attr attrNode = (Attr)attributes.item(k);
          String attrName = attrNode.getNodeName();
          if (attrName.startsWith("xmlns:")
            && attrNode.getNodeValue().equals(namespaceURI))
            return attrName.substring(6);
        }
      }
      {
        Node parentNode = element.getParentNode();
        if (parentNode == null || parentNode.getNodeType() != Node.ELEMENT_NODE)
          return null;
        element = (Element)parentNode;
      }
    }
  }

  // Return the set of explicit prefixes in the context of element.
  private static final Set explicitPrefixSet(Element element) {
    if (element == null)
      throw new IllegalArgumentException();
    Set prefixSet = new HashSet();
    while (true) {
      {
        NamedNodeMap attributes = element.getAttributes();
        for (int k = 0; k < attributes.getLength(); k++) {
          Attr attrNode = (Attr)attributes.item(k);
          String attrName = attrNode.getNodeName();
          if (attrName.startsWith("xmlns:"))
            prefixSet.add(attrName.substring(6));
        }
      }
      {
        Node parentNode = element.getParentNode();
        if (parentNode == null || parentNode.getNodeType() != Node.ELEMENT_NODE)
          break;
        element = (Element)parentNode;
      }
    }
    return prefixSet;
  }

  // Return a sorted list of prefixes (implicit and explicit)
  //   in the context of element.
  public static final String[] prefixes(final Element element) {
    Set prefixSet = explicitPrefixSet(element);
    prefixSet.add("");
    prefixSet.add("xml");
    String[] prefixes = new String[prefixSet.size()];
    prefixSet.toArray(prefixes);
    Arrays.sort(prefixes);
    return prefixes;
  }
}
