/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.util.dom;

import java.util.Random;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Map;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Stack;
import java.io.IOException;

import org.w3c.dom.Node;
import org.w3c.dom.Text;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Attr;
import org.w3c.dom.Element;
import org.w3c.dom.Document;

import org.xmloperator.framework.document.namespace.UndefinedNamespaceException;
import org.xmloperator.framework.document.parse.Dereferencer;
import org.xmloperator.util.io.URI;
import org.xmloperator.util.dom.NamespaceUtils;
import org.xmloperator.util.dom.XmlBaseUtils;
import org.xmloperator.util.dom.Cleaner;
import org.xmloperator.util.dom.ElementBrowser;

// To transform a RELAX NG schema document from full syntax to simple syntax.
public final class RelaxNGSimplifier {

  public static final class RelaxNGSimplificationException extends Exception {

    public RelaxNGSimplificationException(String message) {
      super(message);
    }
  }

  private static final String RELAX_NG_URI =
    "http://relaxng.org/ns/structure/1.0";
  public static int SIMPLIFICATION_TYPE_COMPLETE = 0;
    // For strict simple syntax.
  public static int SIMPLIFICATION_TYPE_GROUP_WITH_MORE_THAN_TWO_CHILDREN = 1;
    // For using to drive editing.
  public static int SIMPLIFICATION_TYPE_REF_RESOLUTION_ONLY = 2;
    // For browsing purpose.
  private Document document;
  private String absoluteURIReference;
  private Dereferencer dereferencer;
  //private boolean mayAGroupHaveMoreThanTwoChildren;
  private int simplificationType; // = SIMPLIFICATION_TYPE_*
  private int phase = 0;
  private Map fromNameTo_define_element = null;

  // pre-condition : the document is a RELAX NG 1.0 full syntax instance.
  // post-condition : the document is a RELAX NG 1.0 simple syntax instance
  //   (in case of full simplification).
  // Some transformation rules are not yet applied (see "Not conformant").
  public static final void simplify(
      Document document, String absoluteURIReference, Dereferencer dereferencer,
      int simplificationType)
      throws RelaxNGSimplificationException {
    new RelaxNGSimplifier(
        document, absoluteURIReference, dereferencer, simplificationType).
        simplify();
  }

  private RelaxNGSimplifier(Document document, String absoluteURIReference,
    Dereferencer dereferencer, int simplificationType) {
    if (document == null || absoluteURIReference == null
        || dereferencer == null)
      throw new IllegalArgumentException();
    this.document = document;
    this.absoluteURIReference = absoluteURIReference;
    this.dereferencer = dereferencer;
    this.simplificationType = simplificationType;
  }

  private void error() throws RelaxNGSimplificationException {
    error("");
  }

  private void error(String message) throws RelaxNGSimplificationException {
    if (message.length() != 0)
      message = " : " + message;
    throw new RelaxNGSimplificationException(
      "exception during simplification phase " + phase + message);
  }

  private void setPhase(int phase) {
    this.phase = phase;
    //System.out.println("phase " + phase);
  }

  private void simplify()
    throws RelaxNGSimplificationException {
    Element rootElement = document.getDocumentElement();
    if (rootElement == null)
      error("There is no document element");
    processAsDataModel_1(rootElement);
    processAsDataModel_2(rootElement);
    processAsDataModel_3(rootElement);
    setPhase(1);
    List temporaryList = new ArrayList();
    processAnnotations_1(rootElement, temporaryList);
    setPhase(2);
    processWhitespace(rootElement);
    if (simplificationType < SIMPLIFICATION_TYPE_REF_RESOLUTION_ONLY) {
      setPhase(3);
      process_datatypeLibrary_attribute_1(rootElement);
      process_datatypeLibrary_attribute_2(rootElement);
      setPhase(4);
      process_type_attributeOf_value_element(rootElement);
    }
    setPhase(5);
    process_href_attribute(rootElement, absoluteURIReference);
    setPhase(6);
    {
      Stack uRIStack = new Stack();
      rootElement = process_externalRef_element(rootElement, uRIStack);
      setPhase(7);
      rootElement =
          process_include_element(rootElement, uRIStack, temporaryList);
    }
    processAnnotations_2(rootElement);
    if (simplificationType < SIMPLIFICATION_TYPE_REF_RESOLUTION_ONLY) {
      setPhase(8);
      process_name_attributeOf_element_and_attribute_elements(rootElement);
      setPhase(9);
      process_ns_attribute_1(rootElement);
      process_ns_attribute_2(rootElement);
    }
    setPhase(10);
    processQNames(rootElement);
    setPhase(11);
    process_div_element(rootElement);
    if (simplificationType < SIMPLIFICATION_TYPE_REF_RESOLUTION_ONLY) {
      setPhase(12);
      processNumberOfChildElements_1(rootElement);
      processNumberOfChildElements_2(rootElement);
      setPhase(13);
      rootElement = process_mixed_element(rootElement);
      setPhase(14);
      rootElement = process_optional_element(rootElement);
      setPhase(15);
      rootElement = process_zeroOrMore_element(rootElement);
    }
    setPhase(16);
    checkConstraints(rootElement);
    setPhase(17);
    process_combine_attribute(rootElement);
    setPhase(18);
    rootElement = process_grammar_element_1();
    {
      final Map from_grammar_toScope = new HashMap();
      process_grammar_element_2(from_grammar_toScope, null, rootElement);
      process_grammar_element_3(from_grammar_toScope, null, null, rootElement);
      process_grammar_element_4(from_grammar_toScope);
    }
    process_grammar_element_5(rootElement);
    if (simplificationType < SIMPLIFICATION_TYPE_REF_RESOLUTION_ONLY) {
      setPhase(19);
      fromNameTo_define_element = new HashMap();
      collectFromNameTo_define_element();
      removeUnreachable_define_elements();
      process_define_and_ref_elements_1(rootElement);
      process_define_and_ref_elements_2();
      process_define_and_ref_elements_3();
      setPhase(20);
      process_notAllowed_element(rootElement);
      removeUnreachable_define_elements();
      setPhase(21);
      process_empty_element(rootElement);
    }
  }

  // Replace the entity references by copies of their childs.
  private void processAsDataModel_1(Element element)
    throws RelaxNGSimplificationException {
    Node childNode = element.getFirstChild();
    while (childNode != null) {
      Node nextChildNode = childNode.getNextSibling();
      switch (childNode.getNodeType()) {
        case Node.ELEMENT_NODE:
          processAsDataModel_1((Element)childNode);
          break;
        case Node.ENTITY_REFERENCE_NODE:
          Node entityNode = childNode.getFirstChild();
          if (entityNode != null)
            nextChildNode = entityNode;
          while (entityNode != null) {
            element.insertBefore(entityNode.cloneNode(true), childNode);
            entityNode = entityNode.getNextSibling();
          }
          element.removeChild(childNode);
          break;
      }
      childNode = nextChildNode;
    }
  }

  // Remove cdata sections, comments and processing instructions.
  private void processAsDataModel_2(Element element)
    throws RelaxNGSimplificationException {
    Node childNode = element.getFirstChild();
    while (childNode != null) {
      Node nextChildNode = childNode.getNextSibling();
      switch (childNode.getNodeType()) {
        case Node.TEXT_NODE:
          break;
        case Node.ELEMENT_NODE:
          processAsDataModel_2((Element)childNode);
          break;
        case Node.CDATA_SECTION_NODE:
        case Node.COMMENT_NODE:
        case Node.PROCESSING_INSTRUCTION_NODE:
          element.removeChild(childNode);
          break;
        default:
          error();
      }
      childNode = nextChildNode;
    }
  }

  // Join consecutive text nodes
  private void processAsDataModel_3(Element element)
    throws RelaxNGSimplificationException {
    Node childNode = element.getFirstChild();
    while (childNode != null) {
      Node nextChildNode = childNode.getNextSibling();
      switch (childNode.getNodeType()) {
        case Node.TEXT_NODE:
          if (nextChildNode != null
            && nextChildNode.getNodeType() == Node.TEXT_NODE) {
            ((Text)childNode).appendData(nextChildNode.getNodeValue());
            element.removeChild(nextChildNode);
            nextChildNode = childNode;
          }
          break;
        case Node.ELEMENT_NODE:
          processAsDataModel_3((Element)childNode);
          break;
        default:
          error();
      }
      childNode = nextChildNode;
    }
  }

  // RELAX NG 1.0 specification  4.1 (1/2)
  // post-condition : there is no more foreign attributes and elements.
  // note : "xml:base" attributes are removed during a second step.
  // pre-condition : temporaryList has any content (don't care).
  // post-condition : temporaryList has any content (don't care).
  private Element processAnnotations_1(Element element, List temporaryList)
    throws RelaxNGSimplificationException {
    if (!this.RELAX_NG_URI.equals(element.getNamespaceURI())) {
      Node parentNode = element.getParentNode();
      if (parentNode.getNodeType() != Node.ELEMENT_NODE)
        error("The document element is a foreign element");
      parentNode.removeChild(element);
      return null;
    }
    {
      temporaryList.clear(); // it's used has a list of attributes to remove.
      NamedNodeMap attributes = element.getAttributes();
      for (int k = 0; k < attributes.getLength(); k++) {
        Attr attr = (Attr)attributes.item(k);
        String namespaceURI = attr.getNamespaceURI();
        if (namespaceURI != null
          && !namespaceURI.equals("http://www.w3.org/2000/xmlns/")
          && !attr.getName().equals("xml:base"))
          temporaryList.add(attr);
      }
      for (int k = 0; k < temporaryList.size(); k++)
        element.removeAttributeNode((Attr)temporaryList.get(k));
    }
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        Element nextChildElement =
          ElementBrowser.getNextSiblingElement(childElement);
        processAnnotations_1(childElement, temporaryList);
        childElement = nextChildElement;
      }
    }
    return element;
  }

  // RELAX NG 1.0 specification  4.1 (2/2)
  // Remove the "xml:base" attributes.
  // post-condition : there is no more foreign attributes and elements.
  private void processAnnotations_2(Element element)
      throws RelaxNGSimplificationException {
    element.removeAttribute("xml:base");
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        processAnnotations_2(childElement);
        childElement = ElementBrowser.getNextSiblingElement(childElement);
      }
    }
  }

  // RELAX NG 1.0 specification  4.2
  // post-condition : there is no more unsignificant whitespace
  private void processWhitespace(Element element)
    throws RelaxNGSimplificationException {
    trimNodeValue(element.getAttributeNode("name"));
    trimNodeValue(element.getAttributeNode("type"));
    trimNodeValue(element.getAttributeNode("combine"));
    String elementName = element.getLocalName();
    boolean isValueElement =
      elementName.equals("value") || elementName.equals("param");
    boolean isNameElement = elementName.equals("name");
    Node childNode = element.getFirstChild();
    while (childNode != null) {
      Node nextChildNode = childNode.getNextSibling();
      switch (childNode.getNodeType()) {
        case Node.ELEMENT_NODE:
          processWhitespace((Element)childNode);
          break;
        case Node.TEXT_NODE: {
          String nodeValue = childNode.getNodeValue();
          if (!isValueElement && Cleaner.isWhitespaceValue(nodeValue))
            element.removeChild(childNode);
          else if (isNameElement)
            childNode.setNodeValue(nodeValue.trim());
          break;
        }
      }
      childNode = nextChildNode;
    }
  }

  private static final void trimNodeValue(Node node) {
    if (node != null)
      node.setNodeValue(node.getNodeValue().trim());
  }

  // RELAX NG 1.0 specification  4.3 (1/2)
  // Not conformant : URI with disallowed characters are refused.
  private void process_datatypeLibrary_attribute_1(Element element)
    throws RelaxNGSimplificationException {
    {
      String datatypeLibrary_attributeValue =
        element.getAttribute("datatypeLibrary");
      if (datatypeLibrary_attributeValue.length() != 0) {
        try {
          new URI(datatypeLibrary_attributeValue);
        }
        catch (URI.MalformedURIException exception) {
          error("Illegal data library URI \""
            + datatypeLibrary_attributeValue + "\"");
        }
      }
      else if (is_data_or_value_element(element)
        && element.getAttributeNode("datatypeLibrary") == null)
        element.setAttribute("datatypeLibrary",
          nearestAncestor_dataLibrary_value(element));
    }
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        process_datatypeLibrary_attribute_1(childElement);
        childElement = ElementBrowser.getNextSiblingElement(childElement);
      }
    }
  }

  private static final boolean is_data_or_value_element(Element element) {
    String elementName = element.getLocalName();
    return elementName.equals("data") || elementName.equals("value");
  }

  private static final String nearestAncestor_dataLibrary_value(
    Element element) {
    while (true) {
      Node parentNode = element.getParentNode();
      if (parentNode.getNodeType() != Node.ELEMENT_NODE)
        return "";
      element = (Element)parentNode;
      Attr attributeNode = element.getAttributeNode("datatypeLibrary");
      if (attributeNode != null)
        return attributeNode.getNodeValue();
    }
  }

  // RELAX NG 1.0 specification  4.3 (2/2)
  // post-condition : there is no more "datatypeLibrary" attribute on an element
  //   other than "data" or "value".
  private void process_datatypeLibrary_attribute_2(Element element)
    throws RelaxNGSimplificationException {
    if (!is_data_or_value_element(element))
      element.removeAttribute("datatypeLibrary");
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        process_datatypeLibrary_attribute_2(childElement);
        childElement = ElementBrowser.getNextSiblingElement(childElement);
      }
    }
  }

  // RELAX NG 1.0 specification  4.4
  // post-condition : every "value" element has a "type" attribute.
  private void process_type_attributeOf_value_element(Element element)
    throws RelaxNGSimplificationException {
    if (element.getLocalName().equals("value")
      && element.getAttributeNode("type") == null) {
      element.setAttribute("type", "token");
      element.setAttribute("datatypeLibrary", "");
    }
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        process_type_attributeOf_value_element(childElement);
        childElement = ElementBrowser.getNextSiblingElement(childElement);
      }
    }
  }

  // RELAX NG 1.0 specification  4.5
  private void process_href_attribute(Element element, String baseURI)
    throws RelaxNGSimplificationException {
    {
      String elementName = element.getLocalName();
      if (elementName.equals("externalRef") || elementName.equals("include")) {
        String localBaseURI = XmlBaseUtils.getBaseURI(baseURI, element);
        String hrefValue = element.getAttribute("href");
        String absoluteURIReference = resolveSystemId(localBaseURI, hrefValue);
        if (absoluteURIReference == null)
          error("Illegal URI reference \"" + hrefValue + "\"");
        element.setAttribute("href", absoluteURIReference);
      }
    }
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        process_href_attribute(childElement, baseURI);
        childElement = ElementBrowser.getNextSiblingElement(childElement);
      }
    }
  }

  private static final String resolveSystemId(String baseSystemId,
    String systemId) {
    // systemId is supposed to be an absolute URI.
    try {
      return new URI(systemId).toString();
    }
    catch (Exception exception) {
    }
    // systemId is supposed to be a relative URI.
    if (baseSystemId != null) try {
      URI base = new URI(baseSystemId);
      return new URI(base, systemId).toString();
    }
    catch (Exception exception) {
    }
    // systemId is not recognized.
    return null;
  }

  private Element dereferenceURIReference(String absoluteURIReference)
      throws RelaxNGSimplificationException {
    Element externalElement = null;
    try {
      externalElement = dereferencer.constructElementUsingURIReference(
          null, absoluteURIReference);
    }
    catch (IOException exception) {
    }
    if (externalElement == null)
      error("Cannot dereference \"" + absoluteURIReference + "\"");
    externalElement =
        processUntilPhase5Included(externalElement, absoluteURIReference);
    if (externalElement == null)
      error("The element constructed using \"" + absoluteURIReference
        + "\" is a foreign element");
    externalElement = (Element)document.importNode(externalElement, true);
    if (externalElement.getAttributeNode("xml:base") == null)
      externalElement.setAttribute("xml:base", absoluteURIReference);
    return externalElement;
  }

  private Element processUntilPhase5Included(Element element, String baseURI)
      throws RelaxNGSimplificationException {
    processAsDataModel_1(element);
    processAsDataModel_2(element);
    processAsDataModel_3(element);
    element = processAnnotations_1(element, new ArrayList());
    if (element == null)
      return null;
    processWhitespace(element);
    if (simplificationType < SIMPLIFICATION_TYPE_REF_RESOLUTION_ONLY) {
      process_datatypeLibrary_attribute_1(element);
      process_datatypeLibrary_attribute_2(element);
      process_type_attributeOf_value_element(element);
    }
    process_href_attribute(element, baseURI);
    return element;
  }

  // RELAX NG 1.0 specification  4.6
  // Return the element (replaced or not replaced).
  private Element process_externalRef_element(Element element, Stack uRIStack)
    throws RelaxNGSimplificationException {
  {
    Element childElement = ElementBrowser.getFirstChildElement(element);
    while (childElement != null) {
      Element nextChildElement =
        ElementBrowser.getNextSiblingElement(childElement);
      process_externalRef_element(childElement, uRIStack);
      childElement = nextChildElement;
    }
  }
  if (element.getLocalName().equals("externalRef")) {
    String absoluteURIReference = element.getAttribute("href");
    if (uRIStack.contains(absoluteURIReference))
      error("Circular externalRef \"" + absoluteURIReference + "\"");
    Element externalElement = dereferenceURIReference(absoluteURIReference);
    uRIStack.push(absoluteURIReference);
    externalElement = process_externalRef_element(externalElement, uRIStack);
    uRIStack.pop();
    if (element.getAttributeNode("ns") != null
        && externalElement.getAttributeNode("ns") == null)
      externalElement.setAttribute("ns", element.getAttribute("ns"));
    Node parentNode = element.getParentNode();
    parentNode.replaceChild(externalElement, element);
    element = externalElement;
  }
  return element;
  }

  // RELAX NG 1.0 specification  4.7
  // Return the element (replaced or not replaced).
  // pre-condition : list has any content (don't care).
  // post-condition : list has any content (don't care).
  private Element process_include_element(Element element, Stack uRIStack,
      List temporaryList)
    throws RelaxNGSimplificationException {
  {
    Element childElement = ElementBrowser.getFirstChildElement(element);
    while (childElement != null) {
      Element nextChildElement =
        ElementBrowser.getNextSiblingElement(childElement);
      process_include_element(childElement, uRIStack, temporaryList);
      childElement = nextChildElement;
    }
  }
  if (element.getLocalName().equals("include")) {
    String absoluteURIReference = element.getAttribute("href");
    if (uRIStack.contains(absoluteURIReference))
      error("Circular include \"" + absoluteURIReference + "\"");
    Element externalElement = dereferenceURIReference(absoluteURIReference);
    if (!externalElement.getLocalName().equals("grammar"))
      error("\"" + absoluteURIReference
        + "\" doesn't refer to a grammar element");
    externalElement = process_externalRef_element(externalElement, new Stack());
    uRIStack.push(absoluteURIReference);
    externalElement =
        process_include_element(externalElement, uRIStack, temporaryList);
    uRIStack.pop();
    if (has_start_component(element)) {
      if (!has_start_component(externalElement))
        error("The grammar element at \"" + absoluteURIReference
          + " has no start component");
      remove_start_components(externalElement);
    }
    temporaryList.clear(); // It's used as a list of "define" components.
    add_define_components(temporaryList, element);
    for (int k = 0; k < temporaryList.size(); k++) {
      Element define_element = (Element)temporaryList.get(k);
      String name = define_element.getAttribute("name");
      if (!has_define_component(externalElement, name))
        error("The grammar element at \"" + absoluteURIReference
          + " has no define component named \"" + name + "\"");
      remove_define_components(externalElement, name);
    }
    Element div_element1 = document.createElementNS(this.RELAX_NG_URI, "div");
    copyAttributesExcept_href(element, div_element1); // not really useful
    Element div_element2 = document.createElementNS(this.RELAX_NG_URI, "div");
    copyAttributesExcept_href(externalElement, div_element2); // not really useful
    moveChildren(externalElement, div_element2);
    Node parentNode = element.getParentNode();
    parentNode.replaceChild(div_element1, element);
    div_element1.appendChild(div_element2);
    moveChildren(element, div_element1);
    element = div_element1;
  }
  return element;
  }

  private static final boolean has_start_component(Element element) {
    Element childElement = ElementBrowser.getFirstChildElement(element);
    while (childElement != null) {
      String childName = childElement.getLocalName();
      if (childName.equals("start")
          || (childName.equals("div") && has_start_component(childElement)))
        return true;
      childElement = ElementBrowser.getNextSiblingElement(childElement);
    }
    return false;
  }

  private static final void remove_start_components(Element element) {
    Element childElement = ElementBrowser.getFirstChildElement(element);
    while (childElement != null) {
      Element nextChildElement =
          ElementBrowser.getNextSiblingElement(childElement);
      String childName = childElement.getLocalName();
      if (childName.equals("start"))
        element.removeChild(childElement);
      else if (childName.equals("div"))
        remove_start_components(childElement);
      childElement = nextChildElement;
    }
  }

  private static final void add_define_components(List define_component_list,
      Element element) {
    Element childElement = ElementBrowser.getFirstChildElement(element);
    while (childElement != null) {
      String childName = childElement.getLocalName();
      if (childName.equals("define"))
        define_component_list.add(childElement);
      else if (childName.equals("div"))
        add_define_components(define_component_list, childElement);
      childElement = ElementBrowser.getNextSiblingElement(childElement);
    }
  }

  private static final boolean has_define_component(Element element,
      String name) {
    Element childElement = ElementBrowser.getFirstChildElement(element);
    while (childElement != null) {
      String childName = childElement.getLocalName();
      if ((childName.equals("define")
           && childElement.getAttribute("name").equals(name))
           || (childName.equals("div")
           && has_define_component(childElement, name)))
        return true;
      childElement = ElementBrowser.getNextSiblingElement(childElement);
    }
    return false;
  }

  private static final void remove_define_components(Element element,
      String name) {
    Element childElement = ElementBrowser.getFirstChildElement(element);
    while (childElement != null) {
      Element nextChildElement =
          ElementBrowser.getNextSiblingElement(childElement);
      String childName = childElement.getLocalName();
      if (childName.equals("define")
          && childElement.getAttribute("name").equals(name))
        element.removeChild(childElement);
      else if (childName.equals("div"))
        remove_define_components(childElement, name);
      childElement = nextChildElement;
    }
  }

  private static final void copyAttributesExcept_href(Element sourceElement,
      Element destinationElement) {
    NamedNodeMap attributes = sourceElement.getAttributes();
    for (int k = 0; k < attributes.getLength(); k++) {
      Attr attr = (Attr)attributes.item(k);
      if (!attr.getName().equals("href"))
        destinationElement.setAttributeNode((Attr)attr.cloneNode(true));
    }
  }

  private static final void moveChildren(Element sourceElement,
      Element destinationElement) {
    while (true) {
      Node childNode = sourceElement.getFirstChild();
      if (childNode == null)
        break;
      destinationElement.appendChild(childNode);
      childNode = childNode.getNextSibling();
    }
  }

  // RELAX NG 1.0 specification  4.8
  private void process_name_attributeOf_element_and_attribute_elements(
    Element element) throws RelaxNGSimplificationException {
    {
      String elementName = element.getLocalName();
      boolean is_attribute_element = elementName.equals("attribute");
      if ((elementName.equals("element") || is_attribute_element)
        && element.getAttributeNode("name") != null) {
        String name_value = element.getAttribute("name");
        element.removeAttribute("name");
        if (name_value.length() != 0) {
          Element name_element =
              document.createElementNS(this.RELAX_NG_URI, "name");
          if (is_attribute_element && element.getAttributeNode("ns") == null)
            name_element.setAttribute("ns", "");
          name_element.appendChild(document.createTextNode(name_value));
          element.insertBefore(name_element, element.getFirstChild());
        }
      }
    }
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        process_name_attributeOf_element_and_attribute_elements(childElement);
        childElement = ElementBrowser.getNextSiblingElement(childElement);
      }
    }
  }

  // RELAX NG 1.0 specification  4.9 (1/2)
  private void process_ns_attribute_1(Element element)
    throws RelaxNGSimplificationException {
    if (isNsFamilyElement(element) && element.getAttributeNode("ns") == null)
      element.setAttribute("ns", nearestAncestor_ns_value(element));
    Element childElement = ElementBrowser.getFirstChildElement(element);
    while (childElement != null) {
      process_ns_attribute_1(childElement);
      childElement = ElementBrowser.getNextSiblingElement(childElement);
    }
  }

  private static final boolean isNsFamilyElement(Element element) {
    String elementName = element.getLocalName();
    return elementName.equals("name") || elementName.equals("nsName")
      || elementName.equals("value");
  }

  private static final String nearestAncestor_ns_value(Element element) {
    while (true) {
      Node parentNode = element.getParentNode();
      if (parentNode.getNodeType() != Node.ELEMENT_NODE)
        return "";
      element = (Element)parentNode;
      Attr attributeNode = element.getAttributeNode("ns");
      if (attributeNode != null)
        return attributeNode.getNodeValue();
    }
  }

  // RELAX NG 1.0 specification  4.9 (2/2)
  private void process_ns_attribute_2(Element element)
    throws RelaxNGSimplificationException {
    if (!isNsFamilyElement(element))
      element.removeAttribute("ns");
    Element childElement = ElementBrowser.getFirstChildElement(element);
    while (childElement != null) {
      process_ns_attribute_2(childElement);
      childElement = ElementBrowser.getNextSiblingElement(childElement);
    }
  }

  // RELAX NG 1.0 specification  4.10
  private void processQNames(Element element)
    throws RelaxNGSimplificationException {
    if (element.getLocalName().equals("name")) {
      Node textNode = element.getFirstChild();
      String value = textNode.getNodeValue();
      int pos_separator = value.indexOf(':');
      if (pos_separator >= 0) {
        String prefix = value.substring(0, pos_separator);
        String localName = value.substring(pos_separator + 1);
        if (prefix.length() == 0 || localName.length() == 0)
          error("Invalid QName \"" + value + "\"");
        textNode.setNodeValue(localName);
        String namespace = null;
        try {
          namespace = NamespaceUtils.prefixToNamespace(element, prefix);
        }
        catch (UndefinedNamespaceException exception) {
        }
        if (namespace == null || namespace.length() == 0)
          error("Undeclared namespace for QName \"" + value + "\"");
        element.setAttribute("ns", namespace);
      }
    }
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        processQNames(childElement);
        childElement = ElementBrowser.getNextSiblingElement(childElement);
      }
    }
  }

  // RELAX NG 1.0 specification  4.11
  // post-condition : there is no more "div" element.
  private void process_div_element(Element element)
    throws RelaxNGSimplificationException {
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        Element nextChildElement =
          ElementBrowser.getNextSiblingElement(childElement);
        process_div_element(childElement);
        childElement = nextChildElement;
      }
    }
    if (element.getLocalName().equals("div")) {
      Node parentNode = element.getParentNode();
      if (parentNode.getNodeType() != Node.ELEMENT_NODE)
        error("The document element must not be a \"div\" element");
      replaceByItsChildren(element);
    }
  }

  // replace a node by its child nodes
  private void replaceByItsChildren(Node node) {
    Node parentNode = node.getParentNode();
    while (true) {
      Node childNode = node.getFirstChild();
      if (childNode == null)
        break;
      parentNode.insertBefore(childNode, node);
    }
    parentNode.removeChild(node);
  }

  // RELAX NG 1.0 specification  4.12 (1/2)
  // post-condition : every "define" element has zero or one child element.
  private void processNumberOfChildElements_1(Element element)
    throws RelaxNGSimplificationException {
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        processNumberOfChildElements_1(childElement);
        childElement = ElementBrowser.getNextSiblingElement(childElement);
      }
    }
    {
      String elementName = element.getLocalName();
      boolean is_element_element =  elementName.equals("element");
      boolean is_except_element = elementName.equals("except");
      if ((is_element_element
        && ElementBrowser.getChildElementCount(element) > 2)
        || ((is_except_element || elementName.equals("define")
        || elementName.equals("oneOrMore") || elementName.equals("zeroOrMore")
        || elementName.equals("optional") || elementName.equals("list")
        || elementName.equals("mixed"))
        && ElementBrowser.getChildElementCount(element) > 1)) {
        Element group_element =
          document.createElementNS(this.RELAX_NG_URI,
          is_except_element ? "choice" : "group");
        Element childElement = ElementBrowser.getFirstChildElement(element);
        if (is_element_element)
          childElement = ElementBrowser.getNextSiblingElement(childElement);
        while (childElement != null) {
          Element nextChildElement =
            ElementBrowser.getNextSiblingElement(childElement);
          group_element.appendChild(childElement);
          childElement = nextChildElement;
        }
        element.appendChild(group_element);
      }
      else if (elementName.equals("attribute")
        && ElementBrowser.getChildElementCount(element) == 1)
        element.appendChild(
          document.createElementNS(this.RELAX_NG_URI, "text"));
    }
  }

  // RELAX NG 1.0 specification  4.12 (2/2)
  private void processNumberOfChildElements_2(Element element)
    throws RelaxNGSimplificationException {
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        Element nextChildElement =
          ElementBrowser.getNextSiblingElement(childElement);
        processNumberOfChildElements_2(childElement);
        childElement = nextChildElement;
      }
    }
    {
      String elementName = element.getLocalName();
      if (elementName.equals("choice") || elementName.equals("group")
        || elementName.equals("interleave"))
        switch (ElementBrowser.getChildElementCount(element)) {
          case 0:
            error("A \"" + elementName
              + "\" must have one or more child element");
          case 1: {
            element.getParentNode().replaceChild(
              ElementBrowser.getFirstChildElement(element), element);
            break;
          }
          default:
            if (simplificationType
                < SIMPLIFICATION_TYPE_GROUP_WITH_MORE_THAN_TWO_CHILDREN)
              while (ElementBrowser.getChildElementCount(element) > 2) {
                Element groupElement =
                    document.createElementNS(this.RELAX_NG_URI, elementName);
                groupElement.appendChild(
                  ElementBrowser.getFirstChildElement(element));
                groupElement.appendChild(
                  ElementBrowser.getFirstChildElement(element));
                element.insertBefore(groupElement, element.getFirstChild());
              }
        }
    }
  }

  // RELAX NG 1.0 specification  4.13
  // post-condition : there is no more "mixed" element.
  private Element process_mixed_element(Element element)
    throws RelaxNGSimplificationException {
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        Element nextChildElement =
          ElementBrowser.getNextSiblingElement(childElement);
        process_mixed_element(childElement);
        childElement = nextChildElement;
      }
    }
    if (element.getLocalName().equals("mixed")) {
      if (ElementBrowser.getChildElementCount(element) != 1)
        error("A \"mixed\" element must contain one child element");
      Element interleave_element =
          document.createElementNS(this.RELAX_NG_URI, "interleave");
      interleave_element.appendChild(
        ElementBrowser.getFirstChildElement(element));
      interleave_element.appendChild(
          document.createElementNS(this.RELAX_NG_URI, "text"));
      element.getParentNode().replaceChild(interleave_element, element);
      element = interleave_element;
    }
    return element;
  }

  // RELAX NG 1.0 specification  4.14
  // post-condition : there is no more "optional" element.
  private Element process_optional_element(Element element)
    throws RelaxNGSimplificationException {
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        Element nextChildElement =
          ElementBrowser.getNextSiblingElement(childElement);
        process_optional_element(childElement);
        childElement = nextChildElement;
      }
    }
    if (element.getLocalName().equals("optional")) {
      if (ElementBrowser.getChildElementCount(element) != 1)
        error("A \"optional\" element must contain one child element");
      Element choice_element =
          document.createElementNS(this.RELAX_NG_URI, "choice");
      choice_element.appendChild(ElementBrowser.getFirstChildElement(element));
      choice_element.appendChild(
          document.createElementNS(this.RELAX_NG_URI, "empty"));
      element.getParentNode().replaceChild(choice_element, element);
      element = choice_element;
    }
    return element;
  }

  // RELAX NG 1.0 specification  4.15
  // post-condition : there is no more "zeroOrMore" element.
  private Element process_zeroOrMore_element(Element element)
    throws RelaxNGSimplificationException {
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        Element nextChildElement =
          ElementBrowser.getNextSiblingElement(childElement);
        process_zeroOrMore_element(childElement);
        childElement = nextChildElement;
      }
    }
    if (element.getLocalName().equals("zeroOrMore")) {
      if (ElementBrowser.getChildElementCount(element) != 1)
        error("A \"zeroOrMore\" element must contain one child element");
      Element choice_element =
          document.createElementNS(this.RELAX_NG_URI, "choice");
      Element oneOrMore_element =
          document.createElementNS(this.RELAX_NG_URI, "oneOrMore");
      oneOrMore_element.appendChild(
        ElementBrowser.getFirstChildElement(element));
      choice_element.appendChild(oneOrMore_element);
      choice_element.appendChild(
          document.createElementNS(this.RELAX_NG_URI, "empty"));
      element.getParentNode().replaceChild(choice_element, element);
      element = choice_element;
    }
    return element;
  }

  // RELAX NG 1.0 specification  4.16
  // Not conformant :
  //   the use of datatypes by data and value elements is not verified.
  private void checkConstraints(Element element)
    throws RelaxNGSimplificationException {
    {
      String elementName = element.getLocalName();
      if (elementName.equals("except")) {
        Node parentNode = element.getParentNode();
        if (parentNode.getNodeType() != Node.ELEMENT_NODE)
          error("The document element must not be an \"except\" element");
        String parentElementName = parentNode.getLocalName();
        if (parentElementName.equals("anyName")
          && hasDescendantWhoseNameIs(element, "anyName"))
          error(
            "An \"anyName\" element must not have an \"anyName\" descendant");
        else if (parentElementName.equals("nsName")
          && (hasDescendantWhoseNameIs(element, "nsName")
          || hasDescendantWhoseNameIs(element, "anyName")))
          error(
            "An \"nsName\" element must not have an \"anyName\" or \"nsName\" descendant");
      }
      else {
        if (elementName.equals("name")
          && isTheFirstChildOfAn_attribute_elementOrADescendant(element)
          && element.getAttribute("ns").length() == 0) {
          Node contentNode = element.getFirstChild();
          if (contentNode == null
            || contentNode.getNodeType() != Node.TEXT_NODE
            || contentNode.getNodeValue().equals("xmlns"))
            error("Invalid name \"xmlns\"");
        }
        if ((elementName.equals("name") || elementName.equals("nsName"))
          && isTheFirstChildOfAn_attribute_elementOrADescendant(element)
          && element.getAttribute("ns").equals("http://www.w3.org/2000/xmlns"))
            error(
              "Invalid \"ns\" attribute value \"http://www.w3.org/2000/xmlns\"");
      }
    }
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        checkConstraints(childElement);
        childElement = ElementBrowser.getNextSiblingElement(childElement);
      }
    }
  }

  private static final boolean hasDescendantWhoseNameIs(Element element,
    String descendantElementName) {
    Element childElement = ElementBrowser.getFirstChildElement(element);
    while (childElement != null) {
      if (childElement.getLocalName().equals(descendantElementName)
        || hasDescendantWhoseNameIs(childElement, descendantElementName))
        return true;
      childElement = ElementBrowser.getNextSiblingElement(childElement);
    }
    return false;
  }

  private static final boolean
    isTheFirstChildOfAn_attribute_elementOrADescendant(Element element) {
    while (true) {
      Node parentNode = element.getParentNode();
      if (parentNode.getNodeType() != Node.ELEMENT_NODE)
        return false;
      if (element.getPreviousSibling() == null
        && parentNode.getLocalName().equals("attribute"))
        return true;
      element = (Element)parentNode;
    }
  }

  // RELAX NG 1.0 specification  4.17
  // post-condition : there is no more combinable "define" or "start" element.
  private void process_combine_attribute(Element element)
    throws RelaxNGSimplificationException {
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        process_combine_attribute(childElement);
        childElement = ElementBrowser.getNextSiblingElement(childElement);
      }
    }
    if (element.getLocalName().equals("grammar")) {
      combine_define_elements(element);
      combine_start_elements(element);
    }
  }

  private final class Combiner {
    private String combineMode = null;
    private boolean isThereElementWithout_combine_attribute = false;
    private final List elementList = new ArrayList();

    public Combiner() {
    }

    public void addElement(Element element)
      throws RelaxNGSimplificationException {
      String combineMode = element.getAttribute("combine");
      if (combineMode.length() == 0) {
        if (isThereElementWithout_combine_attribute)
          error("Missing combine attribute in " + combineContext(element));
        isThereElementWithout_combine_attribute = true;
      }
      else {
        if (this.combineMode != null
            && !this.combineMode.equals(combineMode))
          error("Incoherent combine attribute value in "
            + combineContext(element));
        this.combineMode = combineMode;
      }
      elementList.add(element);
    }

    private String combineContext(Element element) {
      String context = element.getLocalName();
      if (context.equals("define"))
          context += "(\"" + element.getAttribute("name") + "\")";
      return context;
    }

    public void combine() throws RelaxNGSimplificationException {
      int elementCount = elementList.size();
      if (elementCount <= 1)
        return;
      if (!combineMode.equals("choice") && !combineMode.equals("interleave"))
        error("Invalid combine attribute value : \"" + combineMode + "\"");
      Element groupElement =
          document.createElementNS(this.toString(), combineMode);
      for (int k = 0; k < elementCount; k++) {
        Element element = (Element)elementList.get(k);
        while (true) {
          Element patternNode = ElementBrowser.getFirstChildElement(element);
          if (patternNode == null)
            break;
          groupElement.appendChild(patternNode);
        }
        if (k == 0) {
          element.removeAttribute("combine");
          element.appendChild(groupElement);
        }
        else
          element.getParentNode().removeChild(element);
      }
    }
  }

  // Combine combinable "define" child elements.
  private void combine_define_elements(Element grammar_element)
    throws RelaxNGSimplificationException {
    Map fromNameToCombiner = new HashMap();
    {
      Element childElement =
        ElementBrowser.getFirstChildElement(grammar_element);
      while (childElement != null) {
        if (childElement.getLocalName().equals("define")) {
          String name = childElement.getAttribute("name");
          if (name.length() == 0)
            error("Missing \"name\" attribute in \"define\" element");
          Combiner combiner = (Combiner)fromNameToCombiner.get(name);
          if (combiner == null) {
            combiner = new Combiner();
            fromNameToCombiner.put(name, combiner);
          }
          combiner.addElement(childElement);
        }
        childElement = ElementBrowser.getNextSiblingElement(childElement);
      }
    }
    {
      Iterator iterator = fromNameToCombiner.values().iterator();
      while (iterator.hasNext())
        ((Combiner)iterator.next()).combine();
    }
  }

  // Combine combinable "start" child elements.
  private void combine_start_elements(Element grammar_element)
    throws RelaxNGSimplificationException {
    final Combiner combiner = new Combiner();
    {
      Element childElement =
        ElementBrowser.getFirstChildElement(grammar_element);
      while (childElement != null) {
        if (childElement.getLocalName().equals("start"))
          combiner.addElement(childElement);
        childElement = ElementBrowser.getNextSiblingElement(childElement);
      }
    }
    combiner.combine();
  }

  // RELAX NG 1.0 specification  4.18 (1/5)
  // post-condition : The root element is a "grammar" element
  //   whose first child is a "start" element.
  private Element process_grammar_element_1()
    throws RelaxNGSimplificationException {
    Element topLevel_grammar =
        document.createElementNS(this.RELAX_NG_URI, "grammar");
    topLevel_grammar.setAttribute("xmlns", this.RELAX_NG_URI);
    Element topLevel_start =
        document.createElementNS(this.RELAX_NG_URI, "start");
    topLevel_grammar.appendChild(topLevel_start);
    topLevel_start.appendChild(document.getDocumentElement());
    document.appendChild(topLevel_grammar);
    return topLevel_grammar;
  }

  private final class DefineRefs {
    private Element define_element;
    private final List refList = new ArrayList();

    public DefineRefs(Element define_element) {
      this.define_element = define_element;
    }

    public Element get_define_element() {
      return define_element;
    }

    public void addRef(Element refElement) {
      refList.add(refElement);
    }

    public void renameAndIncludeTo(Set nameSet)
      throws RelaxNGSimplificationException {
      String name = define_element.getAttribute("name");
      if (name.length() == 0)
        error("Missing \"name\" attribute in \"define\" element");
      if (nameSet.contains(name)) {
        do {
          name += '_';
        }
        while (nameSet.contains(name));
        define_element.setAttribute("name", name);
        for (int k = 0; k < refList.size(); k++)
          ((Element)refList.get(k)).setAttribute("name", name);
      }
      nameSet.add(name);
    }
  }

  // RELAX NG 1.0 specification  4.18 (2/5)
  // Collects the "grammar" and "define" elements.
  private void process_grammar_element_2(Map from_grammar_toScope,
    Map fromNameToDefineRefs, Element element)
    throws RelaxNGSimplificationException {
    {
      String elementName = element.getLocalName();
      if (elementName.equals("grammar")) {
        fromNameToDefineRefs = new HashMap();
        from_grammar_toScope.put(element, fromNameToDefineRefs);
      }
      else if (elementName.equals("define")) {
        String name = element.getAttribute("name");
        if (name.length() == 0)
          error("Missing \"name\" attribute in \"define\" element");
        if (fromNameToDefineRefs.containsKey(name))
          error("Duplicate \"define\" name \"" + name + "\"");
        fromNameToDefineRefs.put(name, new DefineRefs(element));
      }
    }
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        process_grammar_element_2(from_grammar_toScope, fromNameToDefineRefs,
          childElement);
        childElement = ElementBrowser.getNextSiblingElement(childElement);
      }
    }
  }

  // RELAX NG 1.0 specification  4.18 (3/5)
  // Collects the "ref" and "parenrRef" elements.
  private void process_grammar_element_3(Map from_grammar_toScope,
    Map fromNameToDefineRefs, Map fromNameToParentDefineRefs, Element element)
    throws RelaxNGSimplificationException {
    {
      String elementName = element.getLocalName();
      if (elementName.equals("grammar")) {
        fromNameToParentDefineRefs = fromNameToDefineRefs;
        fromNameToDefineRefs = (Map)from_grammar_toScope.get(element);
      }
      else if (elementName.equals("ref"))
        addRef(element, fromNameToDefineRefs);
      else if (elementName.equals("parentRef"))
        addRef(element, fromNameToParentDefineRefs);
    }
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        process_grammar_element_3(from_grammar_toScope, fromNameToDefineRefs,
          fromNameToParentDefineRefs, childElement);
        childElement = ElementBrowser.getNextSiblingElement(childElement);
      }
    }
  }

  private void addRef(Element refElement, Map fromNameToDefineRefs)
    throws RelaxNGSimplificationException {
    if (refElement.hasChildNodes())
      error("A \"ref\" or \"parentRef\" element must have no child");
    String name = refElement.getAttribute("name");
    if (name.length() == 0)
      error("Missing \"name\" attribute in a \"ref\" or \"parentRef\" element");
    DefineRefs defineRefs = (DefineRefs)fromNameToDefineRefs.get(name);
    if (defineRefs == null)
      error("Missing \"define\" element with name \"" + name + "\"");
    defineRefs.addRef(refElement);
  }

  // RELAX NG 1.0 specification  4.18 (4/5)
  // post-condition : Every "define" element is a child of the root element.
  private void process_grammar_element_4(Map from_grammar_toScope)
    throws RelaxNGSimplificationException {
    final Element rootElement = document.getDocumentElement();
    final Set nameSet = new HashSet();
    Iterator scopeIterator = from_grammar_toScope.values().iterator();
    while (scopeIterator.hasNext()) {
      Map fromNameToDefineRefs = (Map)scopeIterator.next();
      Iterator defineIterator = fromNameToDefineRefs.values().iterator();
      while (defineIterator.hasNext()) {
        DefineRefs defineRefs = (DefineRefs)defineIterator.next();
        defineRefs.renameAndIncludeTo(nameSet);
        rootElement.appendChild(defineRefs.get_define_element());
      }
    }
  }

  // RELAX NG 1.0 specification  4.18 (5/5)
  // post-condition : the root element is the only "grammar" element.
  // post-condition : there is no more "parentRef" element.
  private void process_grammar_element_5(Element element)
    throws RelaxNGSimplificationException {
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        Element nextChildElement =
          ElementBrowser.getNextSiblingElement(childElement);
        process_grammar_element_5(childElement);
        childElement = nextChildElement;
      }
    }
    {
      String elementName = element.getLocalName();
      if (elementName.equals("grammar")) {
        if (element != document.getDocumentElement()) {
          Element start_element =
            ElementBrowser.getFirstChildElement(element, "start");
          if (start_element == null)
            error("\"start\" element missing in \"grammar\" element");
          if (ElementBrowser.getChildElementCount(start_element) != 1)
            error("A \"start\" element must have one child element");
          element.getParentNode().replaceChild(
            ElementBrowser.getFirstChildElement(start_element), element);
        }
      }
      else if (elementName.equals("parentRef")) {
        Element refElement =
            document.createElementNS(this.RELAX_NG_URI, "ref");
        refElement.setAttribute("name", element.getAttribute("name"));
        element.getParentNode().replaceChild(refElement, element);
      }
    }
  }

  private void collectFromNameTo_define_element()
    throws RelaxNGSimplificationException {
    Element define_element =
      ElementBrowser.getNextSiblingElement(
        ElementBrowser.getFirstChildElement(
        document.getDocumentElement()));
    while (define_element != null) {
      fromNameTo_define_element.put(define_element.getAttribute("name"),
        define_element);
      define_element = ElementBrowser.getNextSiblingElement(define_element);
    }
  }

  private void removeUnreachable_define_elements() {
    final Set reachable_define_elementSet = new HashSet();
    Element start_element =
      ElementBrowser.getFirstChildElement(document.getDocumentElement());
    collectTheReachable_define_elements(reachable_define_elementSet,
      start_element); // process from the start element
    removeUnreachable_define_elements(reachable_define_elementSet);
  }

  private void collectTheReachable_define_elements(
    Set reachable_define_elementSet, Element element) {
    if (element.getLocalName().equals("ref")) {
      Element define_element =
        (Element)fromNameTo_define_element.get(element.getAttribute("name"));
      if (!reachable_define_elementSet.contains(define_element)) {
        reachable_define_elementSet.add(define_element);
        collectTheReachable_define_elements(reachable_define_elementSet,
          define_element);
      }
    }
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        collectTheReachable_define_elements(reachable_define_elementSet,
          childElement);
        childElement = ElementBrowser.getNextSiblingElement(childElement);
      }
    }
  }

  private void removeUnreachable_define_elements(
    Set reachable_define_elementSet) {
    Element grammarElement = document.getDocumentElement();
    Element define_element = ElementBrowser.getNextSiblingElement(
      ElementBrowser.getFirstChildElement(grammarElement));
    while (define_element != null) {
      Element next_define_element =
        ElementBrowser.getNextSiblingElement(define_element);
      if (!reachable_define_elementSet.contains(define_element))
        grammarElement.removeChild(define_element);
      define_element = next_define_element;
    }
  }

  private static final NameGenerator nameGenerator = new NameGenerator();
  private static final class NameGenerator {
    private static final String chars = "abcdefghijklmnopqrstuvwxyz";
    private Random random = new Random(0);

    public NameGenerator() {
    }

    public String newName(String start, Set existingNames) {
      StringBuffer sb = new StringBuffer(start);
      sb.append('_');
      while (true) {
        String name = sb.toString();
        if (!existingNames.contains(name))
          return name;
        sb.append(chars.charAt(random.nextInt(chars.length())));
      }
    }
  }

  // RELAX NG 1.0 specification  4.19 (1/3)
  // post-condition : every "element" element is the child of a "define" element.
  // Return the element by which it is replaced.
  private Element process_define_and_ref_elements_1(Element element)
    throws RelaxNGSimplificationException {
    if (element.getLocalName().equals("element")
      && !((Element)element.getParentNode()).getLocalName().equals("define")) {
      String defineName = nameGenerator.newName(get_element_name(element),
        fromNameTo_define_element.keySet());
      Element define_element =
          document.createElementNS(this.RELAX_NG_URI, "define");
      define_element.setAttribute("name", defineName);
      document.getDocumentElement().appendChild(define_element);
      Element refElement = document.createElementNS(this.RELAX_NG_URI, "ref");
      refElement.setAttribute("name", defineName);
      ((Element)element.getParentNode()).insertBefore(refElement, element);
      define_element.appendChild(element);
        // children of element will be processed.
      fromNameTo_define_element.put(defineName, define_element);
      return refElement;
    }
    else {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        childElement = process_define_and_ref_elements_1(childElement);
        childElement = ElementBrowser.getNextSiblingElement(childElement);
      }
      return element;
    }
  }

  private String get_element_name(Element element_element) {
    Element name_element =
      ElementBrowser.getFirstChildElement(element_element, "name");
    if (name_element == null)
      return "";
    Node name_contentNode = name_element.getFirstChild();
    return name_contentNode != null
      && name_contentNode.getNodeType() == Node.TEXT_NODE ?
      name_contentNode.getNodeValue() : "";
  }

  // RELAX NG 1.0 specification  4.19 (2/3)
  // post-condition : every "ref" element within "start" or "element" element
  //   is expanded.
  private void process_define_and_ref_elements_2()
    throws RelaxNGSimplificationException {
    Element start_element =
      ElementBrowser.getFirstChildElement(document.getDocumentElement());
    processExpandable_ref_elements(new Stack(), start_element);
    Element define_element =
      ElementBrowser.getNextSiblingElement(start_element);
    while (define_element != null) {
      Element patternElement =
        ElementBrowser.getFirstChildElement(define_element);
      if (patternElement == null)
        error("Pattern element missing in \"define\" element");
      if (patternElement.getLocalName().equals("element"))
        processExpandable_ref_elements(new Stack(), patternElement);
      define_element = ElementBrowser.getNextSiblingElement(define_element);
    }
  }

  private void processExpandable_ref_elements(
    Stack expanded_define_elementStack, Element element)
    throws RelaxNGSimplificationException {
    if (element.getLocalName().equals("ref")) {
      Element define_element =
        (Element)fromNameTo_define_element.get(element.getAttribute("name"));
      Element patternElement =
        ElementBrowser.getFirstChildElement(define_element);
      if (patternElement.getLocalName().equals("element"))
        return;
      if (expanded_define_elementStack.contains(define_element))
        error("Loop while expanding ref elements");
      expanded_define_elementStack.push(define_element);
      Element clonedPatternElement = (Element)patternElement.cloneNode(true);
      element.getParentNode().replaceChild(clonedPatternElement, element);
      processExpandable_ref_elements(expanded_define_elementStack,
        clonedPatternElement);
      expanded_define_elementStack.pop();
    }
    else {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        Element nextChildElement =
          ElementBrowser.getNextSiblingElement(childElement);
        processExpandable_ref_elements(expanded_define_elementStack,
          childElement);
        childElement = nextChildElement;
      }
    }
  }

  // RELAX NG 1.0 specification  4.19 (3/3)
  // post-condition : There is no more "define" element whose child is not
  //   an "element" element.
  private void process_define_and_ref_elements_3()
    throws RelaxNGSimplificationException {
    Element grammar_element = document.getDocumentElement();
    Element define_element =
      ElementBrowser.getNextSiblingElement(
      ElementBrowser.getFirstChildElement(grammar_element));
    while (define_element != null) {
      Element next_define_element =
        ElementBrowser.getNextSiblingElement(define_element);
      if (!ElementBrowser.getFirstChildElement(define_element).
        getLocalName().equals("element"))
        grammar_element.removeChild(define_element);
      define_element = next_define_element;
    }
  }

  // RELAX NG 1.0 specification  4.20
  private void process_notAllowed_element(Element element)
    throws RelaxNGSimplificationException {
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        Element nextChildElement =
          ElementBrowser.getNextSiblingElement(childElement);
        process_notAllowed_element(childElement);
        childElement = nextChildElement;
      }
    }
    {
      String elementName = element.getLocalName();
      if (elementName.equals("attribute") || elementName.equals("list")
        || elementName.equals("group") || elementName.equals("interleave")
        || elementName.equals("oneOrMore")) {
        if (ElementBrowser.getFirstChildElement(element, "notAllowed") != null)
          element.getParentNode().replaceChild(
            document.createElementNS(this.RELAX_NG_URI, "notAllowed"), element);
      }
      else if (elementName.equals("choice")) {
        removeChildren(element, "notAllowed");
        replaceDegeneratedPattern(element, "notAllowed");
      }
      else if (elementName.equals("except")) {
        if (ElementBrowser.getFirstChildElement(element, "notAllowed") != null)
          element.getParentNode().removeChild(element);
      }
    }
  }

  // Remove children of element whose name is childElementName
  private static final void removeChildren(Element element,
    String childElementName) {
    Element childElement = ElementBrowser.getFirstChildElement(element);
    while (childElement != null) {
      Element nextChildElement =
        ElementBrowser.getNextSiblingElement(childElement);
      if (childElement.getLocalName().equals(childElementName))
        element.removeChild(childElement);
      childElement = nextChildElement;
    }
  }

  private void replaceDegeneratedPattern(Element element, String elementName) {
    switch (ElementBrowser.getChildElementCount(element)) {
      case 0:
        element.getParentNode().replaceChild(
          document.createElementNS(this.RELAX_NG_URI, elementName), element);
        break;
      case 1:
        element.getParentNode().replaceChild(
          ElementBrowser.getFirstChildElement(element), element);
        break;
    }
  }

  // RELAX NG 1.0 specification  4.21
  private void process_empty_element(Element element)
    throws RelaxNGSimplificationException {
    {
      Element childElement = ElementBrowser.getFirstChildElement(element);
      while (childElement != null) {
        process_empty_element(childElement);
        childElement = ElementBrowser.getNextSiblingElement(childElement);
      }
    }
    {
      String elementName = element.getLocalName();
      if (elementName.equals("group") || elementName.equals("interleave")) {
        removeChildren(element, "empty");
        replaceDegeneratedPattern(element, "empty");
      }
      else if (elementName.equals("choice")) {
        if (ElementBrowser.getFirstChildElement(element, "empty") != null) {
          removeChildren(element, "empty");
          if (ElementBrowser.getChildElementCount(element) == 0)
            element.getParentNode().replaceChild(
              document.createElementNS(this.RELAX_NG_URI, "empty"), element);
          else
            element.insertBefore(
                document.createElementNS(this.RELAX_NG_URI, "empty"),
                element.getFirstChild());
        }
      }
      else if (elementName.equals("oneOrMore")) {
        if (ElementBrowser.getFirstChildElement(element, "empty") != null)
          element.getParentNode().replaceChild(
            document.createElementNS(this.RELAX_NG_URI, "empty"), element);
      }
    }
  }

  /*
  private static final String elementToString(Element element) {
    StringBuffer sb = new StringBuffer();
    Node parentNode = element.getParentNode();
    if (parentNode.getNodeType() == Node.ELEMENT_NODE) {
      sb.append(elementToString((Element)parentNode));
      sb.append('/');
    }
    sb.append(element.getLocalName());
    sb.append('(');
    sb.append(element.getAttribute("name"));
    sb.append(')');
    return sb.toString();
  }
  */
}
